import datetime
import logging
from subprocess import Popen
from time import sleep
from unittest import expectedFailure, skip, mock

import os
from django.test import TransactionTestCase
from django.test.testcases import TestCase
from django.utils.timezone import now
from random import randint

from django_clickhouse.database import connections
from django_clickhouse.migrations import migrate_app
from django_clickhouse.storages import RedisStorage
from django_clickhouse.tasks import sync_clickhouse_model, clickhouse_auto_sync
from django_clickhouse.utils import int_ranges
from tests.clickhouse_models import ClickHouseTestModel, ClickHouseCollapseTestModel, ClickHouseMultiTestModel
from tests.models import TestModel

logger = logging.getLogger('django-clickhouse')


class SyncTest(TransactionTestCase):
    def setUp(self):
        self.db = ClickHouseCollapseTestModel.get_database()
        self.db.drop_database()
        self.db.create_database()
        migrate_app('tests', 'default')
        ClickHouseTestModel.get_storage().flush()

    def test_simple(self):
        obj = TestModel.objects.create(value=1, created=now(), created_date=datetime.date.today())
        ClickHouseTestModel.sync_batch_from_storage()

        synced_data = list(ClickHouseTestModel.objects.all())
        self.assertEqual(1, len(synced_data))
        self.assertEqual(obj.created_date, synced_data[0].created_date)
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

    def test_collapsing_update_by_final(self):
        obj = TestModel.objects.create(value=1, created=now(), created_date=datetime.date.today())
        obj.value = 2
        obj.save()
        ClickHouseCollapseTestModel.sync_batch_from_storage()

        # insert and update came before sync. Only one item will be inserted
        synced_data = list(ClickHouseCollapseTestModel.objects.all())
        self.assertEqual(1, len(synced_data))
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

        obj.value = 3
        obj.save()
        ClickHouseCollapseTestModel.sync_batch_from_storage()

        synced_data = list(self.db.select('SELECT * FROM $table FINAL', model_class=ClickHouseCollapseTestModel))
        self.assertGreaterEqual(len(synced_data), 1)
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

    def test_collapsing_update_by_version(self):
        ClickHouseCollapseTestModel.engine.version_col = 'version'

        obj = TestModel.objects.create(value=1, created=now(), created_date=datetime.date.today())
        obj.value = 2
        obj.save()
        ClickHouseCollapseTestModel.sync_batch_from_storage()

        # insert and update came before sync. Only one item will be inserted
        synced_data = list(ClickHouseCollapseTestModel.objects.all())
        self.assertEqual(1, len(synced_data))
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

        obj.value = 3
        obj.save()
        ClickHouseCollapseTestModel.sync_batch_from_storage()

        synced_data = list(self.db.select('SELECT * FROM $table FINAL', model_class=ClickHouseCollapseTestModel))
        self.assertGreaterEqual(len(synced_data), 1)
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

        ClickHouseCollapseTestModel.engine.version_col = None

    @expectedFailure
    def test_collapsing_delete(self):
        obj = TestModel.objects.create(value=1, created_date=datetime.date.today())
        ClickHouseCollapseTestModel.sync_batch_from_storage()
        obj.delete()
        ClickHouseCollapseTestModel.sync_batch_from_storage()

        # sync_batch_from_storage uses FINAL, so data would be collapsed by now
        synced_data = list(ClickHouseCollapseTestModel.objects.all())
        self.assertEqual(0, len(synced_data))

    def test_multi_model(self):
        obj = TestModel.objects.create(value=1, created=now(), created_date=datetime.date.today())
        obj.value = 2
        obj.save()
        ClickHouseMultiTestModel.sync_batch_from_storage()

        synced_data = list(ClickHouseTestModel.objects.all())
        self.assertEqual(1, len(synced_data))
        self.assertEqual(obj.created_date, synced_data[0].created_date)
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

        # sync_batch_from_storage uses FINAL, so data would be collapsed by now
        synced_data = list(ClickHouseCollapseTestModel.objects.all())
        self.assertEqual(1, len(synced_data))
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)

        obj.value = 3
        obj.save()
        ClickHouseMultiTestModel.sync_batch_from_storage()

        synced_data = list(self.db.select('SELECT * FROM $table FINAL', model_class=ClickHouseCollapseTestModel))
        self.assertGreaterEqual(len(synced_data), 1)
        self.assertEqual(obj.value, synced_data[0].value)
        self.assertEqual(obj.id, synced_data[0].id)


class KillTest(TransactionTestCase):
    TEST_TIME = 60
    maxDiff = None

    def setUp(self):
        ClickHouseTestModel.get_storage().flush()
        connections['default'].drop_database()
        connections['default'].create_database()
        migrate_app('tests', 'default')

        # Disable sync for not interesting models
        ClickHouseMultiTestModel.sync_enabled = False
        ClickHouseTestModel.sync_enabled = False

    def tearDown(self):
        # Disable sync for not interesting models
        ClickHouseMultiTestModel.sync_enabled = True
        ClickHouseTestModel.sync_enabled = True

    def _check_data(self):
        logger.debug('django-clickhouse: syncing left test data')

        # Sync all data that is not synced
        # Data is expected to be in test_db, not default. So we need to call subprocess
        # in order everything works correctly
        import_key = ClickHouseCollapseTestModel.get_import_key()
        storage = ClickHouseCollapseTestModel.get_storage()
        sync_left = storage.operations_count(import_key)
        while sync_left:
            logger.debug('django-clickhouse: final sync (%d left)' % sync_left)
            self.sync_iteration(False)
            sync_left = storage.operations_count(import_key)

        logger.debug('django_clickhouse: sync finished')

        ch_data = list(connections['default'].select_tuples('SELECT * FROM $table FINAL ORDER BY id',
                                                            model_class=ClickHouseCollapseTestModel))
        logger.debug('django_clickhouse: got clickhouse data')

        pg_data = list(TestModel.objects.all().order_by('id'))
        logger.debug('django_clickhouse: got postgres data')

        if len(pg_data) != len(ch_data):
            absent_ids = set(item.id for item in pg_data) - set(item.id for item in ch_data)
            logger.debug('django_clickhouse: absent ranges: %s (min: %d, max: %d)'
                         % (','.join(('(%d, %d)' % r) for r in int_ranges(absent_ids)),
                            min(item.id for item in pg_data), max(item.id for item in pg_data)))

        self.assertEqual(len(pg_data), len(ch_data))
        for pg_item, ch_item in zip(pg_data, ch_data):
            self.assertEqual(ch_item.id, pg_item.id)
            self.assertEqual(ch_item.value, pg_item.value)

    @classmethod
    def sync_iteration(cls, kill=True):
        test_script = os.path.join(os.path.dirname(__file__), 'kill_test_sub_process.py')
        if kill:
            args = ['--test-time', str(cls.TEST_TIME)]
        else:
            args = ['--once', 'true']
        p_sync = Popen(['python3', test_script, 'sync'] + args)

        if kill:
            sleep(randint(0, 5))
            logger.debug('django-clickhouse: test killing: %d' % p_sync.pid)
            p_sync.kill()
        else:
            p_sync.wait()

    def test_kills(self):
        test_script = os.path.join(os.path.dirname(__file__), 'kill_test_sub_process.py')
        p_create = Popen(['python3', test_script, 'create', '--test-time', str(self.TEST_TIME)])

        # Updates must be slower than inserts, or they will do nothing
        p_update = Popen(['python3', test_script, 'update', '--test-time', str(self.TEST_TIME), '--batch-size', '500'])

        start = now()
        while (now() - start).total_seconds() < self.TEST_TIME:
            self.sync_iteration()

        p_create.wait()
        p_update.wait()

        self._check_data()


@mock.patch.object(ClickHouseTestModel, 'sync_batch_from_storage')
class SyncClickHouseModelTest(TestCase):
    def test_model_as_class(self, sync_mock):
        sync_clickhouse_model(ClickHouseTestModel)
        sync_mock.assert_called()

    def test_model_as_string(self, sync_mock):
        sync_clickhouse_model('tests.clickhouse_models.ClickHouseTestModel')
        sync_mock.assert_called()

    @mock.patch.object(RedisStorage, 'set_last_sync_time')
    def test_last_sync_time_called(self, storage_mock, _):
        sync_clickhouse_model(ClickHouseTestModel)
        storage_mock.assert_called()
        self.assertEqual(2, len(storage_mock.call_args))
        self.assertEqual(storage_mock.call_args[0][0], 'ClickHouseTestModel')
        self.assertIsInstance(storage_mock.call_args[0][1], datetime.datetime)


@mock.patch.object(sync_clickhouse_model, 'delay')
class ClickHouseAutoSyncTest(TestCase):
    @mock.patch('django_clickhouse.tasks.get_subclasses', return_value=[ClickHouseTestModel])
    @mock.patch.object(ClickHouseTestModel, 'need_sync', return_value=True)
    def test_needs_sync_enabled(self, need_sync_mock, get_subclasses_mock, sync_delay_mock):
        clickhouse_auto_sync()
        sync_delay_mock.assert_called_with('tests.clickhouse_models.ClickHouseTestModel')

    @mock.patch('django_clickhouse.tasks.get_subclasses', return_value=[ClickHouseTestModel])
    @mock.patch.object(ClickHouseTestModel, 'need_sync', return_value=False)
    def test_does_not_need_sync(self, need_sync_mock, get_subclasses_mock, sync_delay_mock):
        clickhouse_auto_sync()
        sync_delay_mock.assert_not_called()

    @mock.patch('django_clickhouse.tasks.get_subclasses',
                return_value=[ClickHouseTestModel, ClickHouseCollapseTestModel])
    @mock.patch.object(ClickHouseTestModel, 'need_sync', return_value=True)
    @mock.patch.object(ClickHouseCollapseTestModel, 'need_sync', return_value=True)
    def test_multiple_models(self, need_sync_1_mock, need_sync_2_mock, get_subclasses_mock, sync_delay_mock):
        clickhouse_auto_sync()
        self.assertEqual(2, sync_delay_mock.call_count)


# Used to profile sync execution time. Disabled by default
@skip
class ProfileTest(TransactionTestCase):
    BATCH_SIZE = 10000

    def setUp(self):
        ClickHouseTestModel.get_storage().flush()
        connections['default'].drop_database()
        connections['default'].create_database()
        migrate_app('tests', 'default')

        # Disable sync for not interesting models
        ClickHouseMultiTestModel.sync_enabled = False
        ClickHouseTestModel.sync_enabled = False

        TestModel.objects.bulk_create([
            TestModel(created=now(), created_date='2018-01-01', value=i)
            for i in range(self.BATCH_SIZE)
        ])

    def tearDown(self):
        # Disable sync for not interesting models
        ClickHouseMultiTestModel.sync_enabled = True
        ClickHouseTestModel.sync_enabled = True

    def test_sync(self):
        ClickHouseCollapseTestModel.sync_batch_size = self.BATCH_SIZE
        ClickHouseCollapseTestModel.sync_batch_from_storage()



import datetime

from django.test import TestCase
from django.utils.timezone import now

from tests.clickhouse_models import ClickHouseTestModel


class ClickHouseModelTest(TestCase):
    def setUp(self):
        self.storage = ClickHouseTestModel.get_storage()
        self.storage.flush()

    def test_need_sync(self):
        # sync is disabled by default
        ClickHouseTestModel.sync_enabled = False
        self.assertFalse(ClickHouseTestModel.need_sync())

        # There were no syncs. So it should be done
        ClickHouseTestModel.sync_enabled = True
        self.assertTrue(ClickHouseTestModel.need_sync())

        # Time hasn't passed - no sync
        self.storage.set_last_sync_time(ClickHouseTestModel.get_import_key(), now())
        self.assertFalse(ClickHouseTestModel.need_sync())

        # Time has passed
        sync_delay = ClickHouseTestModel.get_sync_delay()
        self.storage.set_last_sync_time(ClickHouseTestModel.get_import_key(),
                                        now() - datetime.timedelta(seconds=sync_delay + 1))
        self.assertTrue(ClickHouseTestModel.need_sync())



from .configuration import PREFIX


class ConfigurationError(Exception):
    def __init__(self, param_name):
        param_name = PREFIX + param_name
        super(ConfigurationError, self).__init__("Config parameter '%s' is not set properly" % param_name)


class DBAliasError(Exception):
    def __init__(self, alias):
        super(DBAliasError, self).__init__(
            "Database alias `%s` is not found. Check %s parameter" % (alias, PREFIX + 'DATABASES'))


class RedisLockTimeoutError(Exception):
    pass



from django.test import TestCase

from django_clickhouse.configuration import config


class ConfigTest(TestCase):
    def test_default(self):
        self.assertEqual(5, config.SYNC_DELAY)

    def test_value(self):
        self.assertEqual(5000, config.SYNC_BATCH_SIZE)

    def test_not_lib_prop(self):
        with self.assertRaises(AttributeError):
            config.SECRET_KEY



import datetime
from queue import Queue

import pytz
from django.test import TestCase

from django_clickhouse.models import ClickHouseSyncModel
from django_clickhouse.utils import get_tz_offset, format_datetime, lazy_class_import, int_ranges, exec_in_parallel, \
    SingletonMeta


def local_dt_str(dt) -> str:
    """
    Returns string representation of an aware datetime object, localized by adding system_tz_offset()
    :param dt: Datetime to change
    :return: Formatted string
    """
    return (dt + datetime.timedelta(minutes=get_tz_offset())).strftime('%Y-%m-%d %H:%M:%S')


class FormatDateTimeTest(TestCase):
    @staticmethod
    def _get_zone_time(dt):
        """
        На момент написания тестов в РФ было какое-то странное смещение (для Москвы, например +2:30, для Перми +4:03)
        :param dt: Объект datetime.datetime
        :return: Строковый ожидаемый результат
        """
        minute_offset = dt.utcoffset().total_seconds() / 60
        zone_h, zone_m = abs(int(minute_offset / 60)), int(minute_offset % 60)

        return local_dt_str(dt - datetime.timedelta(hours=zone_h, minutes=zone_m))

    def test_conversion(self):
        dt = datetime.datetime(2017, 1, 2, 3, 4, 5)
        self.assertEqual(format_datetime(dt), local_dt_str(dt))
        dt = datetime.datetime(2017, 1, 2, 3, 4, 5, tzinfo=pytz.utc)
        self.assertEqual(format_datetime(dt), local_dt_str(dt))
        dt = datetime.datetime(2017, 1, 2, 3, 4, 5, tzinfo=pytz.timezone('Europe/Moscow'))
        self.assertEqual(format_datetime(dt), self._get_zone_time(dt))
        dt = datetime.datetime(2017, 1, 2, 3, 4, 5, tzinfo=pytz.timezone('Europe/Moscow'))
        offset = int(pytz.timezone('Europe/Moscow').utcoffset(dt).total_seconds() / 60)
        self.assertEqual(format_datetime(dt, timezone_offset=offset),
                         local_dt_str(datetime.datetime(2017, 1, 2, 3, 4, 5) - datetime.timedelta(minutes=offset*2)))

    def test_date_conversion(self):
        dt = datetime.date(2017, 1, 2)
        self.assertEqual(format_datetime(dt), local_dt_str(datetime.datetime(2017, 1, 2, 0, 0, 0)))
        dt = datetime.date(2017, 1, 2)
        self.assertEqual(format_datetime(dt, day_end=True), local_dt_str(datetime.datetime(2017, 1, 2, 23, 59, 59)))
        dt = datetime.date(2017, 1, 2)
        self.assertEqual(format_datetime(dt, day_end=True, timezone_offset=60),
                         local_dt_str(datetime.datetime(2017, 1, 2, 22, 59, 59)))
        dt = datetime.date(2017, 1, 2)
        self.assertEqual(format_datetime(dt, timezone_offset=60), local_dt_str(datetime.datetime(2017, 1, 1, 23, 0, 0)))


class TestLazyClassImport(TestCase):
    def test_str(self):
        self.assertEqual(ClickHouseSyncModel, lazy_class_import('django_clickhouse.models.ClickHouseSyncModel'))

    def test_cls(self):
        self.assertEqual(ClickHouseSyncModel, lazy_class_import(ClickHouseSyncModel))


class TestIntRanges(TestCase):
    def test_simple(self):
        self.assertListEqual([(1, 3), (5, 6), (8, 10)],
                             list(int_ranges([1, 2, 3, 5, 6, 8, 9, 10])))

    def test_empty(self):
        self.assertListEqual([], list(int_ranges([])))

    def test_bounds(self):
        self.assertListEqual([(1, 1), (5, 6), (10, 10)],
                             list(int_ranges([1, 5, 6, 10])))


class TestExecInParallel(TestCase):
    base_classes = []

    def test_exec(self):
        q = Queue()
        for i in range(10):
            q.put(([i], {}))

        res = exec_in_parallel(lambda x: x*x, q, 4)
        self.assertSetEqual({x * x for x in range(10)}, set(res))

    def test_exec_no_count(self):
        q = Queue()
        for i in range(10):
            q.put(([i], {}))

        res = exec_in_parallel(lambda x: x * x, q)
        self.assertSetEqual({x * x for x in range(10)}, set(res))

    def test_exception(self):
        q = Queue()
        for i in range(10):
            q.put(([i], {}))

        def _test_func(x):
            raise TypeError("Exception in thread %d" % x)

        with self.assertRaises(TypeError):
            exec_in_parallel(_test_func, q)


class TestSingletonMeta(TestCase):
    def test_singleton(self):
        class Single(metaclass=SingletonMeta):
            def __init__(self):
                self.test = 1

        a = Single()
        a.test += 1
        b = Single()
        self.assertEqual(a, b)
        self.assertEqual(2, b.test)



import datetime

from django.test import TestCase

from django_clickhouse.serializers import Django2ClickHouseModelSerializer
from tests.clickhouse_models import ClickHouseTestModel
from tests.models import TestModel


class Django2ClickHouseModelSerializerTest(TestCase):
    fixtures = ['test_model']

    def setUp(self):
        self.obj = TestModel.objects.get(pk=1)

    def test_all(self):
        serializer = Django2ClickHouseModelSerializer(ClickHouseTestModel)
        res = serializer.serialize(self.obj)
        self.assertEqual(self.obj.id, res.id)
        self.assertEqual(self.obj.value, res.value)
        self.assertEqual(self.obj.created_date, res.created_date)

    def test_fields(self):
        serializer = Django2ClickHouseModelSerializer(ClickHouseTestModel, fields=('value',))
        res = serializer.serialize(self.obj)
        self.assertEqual(0, res.id)
        self.assertEqual(datetime.date(1970, 1, 1), res.created_date)
        self.assertEqual(self.obj.value, res.value)

    def test_exclude_fields(self):
        serializer = Django2ClickHouseModelSerializer(ClickHouseTestModel, exclude_fields=('created_date',))
        res = serializer.serialize(self.obj)
        self.assertEqual(datetime.date(1970, 1, 1), res.created_date)
        self.assertEqual(self.obj.id, res.id)
        self.assertEqual(self.obj.value, res.value)



from typing import Optional, Iterable, List, Type

from copy import copy
from infi.clickhouse_orm.database import Database
from infi.clickhouse_orm.models import Model as InfiModel
from infi.clickhouse_orm.query import QuerySet as InfiQuerySet, AggregateQuerySet as InfiAggregateQuerySet

from .database import connections


class QuerySet(InfiQuerySet):
    """
    Basic QuerySet to use
    """

    def __init__(self, model_cls: Type[InfiModel], database: Optional[Database] = None) -> None:
        super(QuerySet, self).__init__(model_cls, database)
        self._db_alias = None

    @property
    def _database(self) -> Database:
        # HACK for correct work of all infi.clickhouse-orm methods
        # There are no write QuerySet methods now, so I use for_write=False by default
        return self.get_database(for_write=False)

    @_database.setter
    def _database(self, database: Database) -> None:
        # HACK for correct work of all infi.clickhouse-orm methods
        self._db = database

    def get_database(self, for_write: bool = False) -> Database:
        """
        Gets database to execute query on. Looks for constructor or using() method.
        If nothing was set tries to get database from model class using router.
        :param for_write: Return QuerySet for read or for write.
        :return: Database instance
        """
        if not self._db:
            if self._db_alias:
                self._db = connections[self._db_alias]
            else:
                self._db = self._model_cls.get_database(for_write=for_write)

        return self._db

    def using(self, db_alias: str) -> 'QuerySet':
        """
        Sets database alias to use for this query
        :param db_alias: Database alias name from CLICKHOUSE_DATABASES config option
        :return: None
        """
        qs = copy(self)
        qs._db_alias = db_alias
        qs._db = None  # Previous database should be forgotten
        return qs

    def all(self) -> 'QuerySet':
        """
        Returns all items of queryset
        :return: QuerySet
        """
        return copy(self)

    def create(self, **kwargs):
        """
        Create single item in database
        :return: Created instance
        """
        instance = self._model_cls(**kwargs)
        self.get_database(for_write=True).insert([instance])
        return instance

    def bulk_create(self, model_instances: Iterable[InfiModel], batch_size: int = 1000) -> List[InfiModel]:
        self.get_database(for_write=True).insert(model_instances=model_instances, batch_size=batch_size)
        return list(model_instances)


class AggregateQuerySet(QuerySet, InfiAggregateQuerySet):
    pass



"""
Django command that applies migrations for ClickHouse database
"""
import json

from django.apps import apps as django_apps
from django.core.management import BaseCommand, CommandParser

from ...configuration import config
from ...migrations import migrate_app


class Command(BaseCommand):
    help = 'Migrates ClickHouse databases'
    requires_migrations_checks = False

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('app_label', nargs='?', type=str,
                            help='Django App name to migrate. By default all found apps are migrated.')

        parser.add_argument('migration_number', nargs='?', type=int,
                            help='Migration number in selected django app to migrate to.'
                                 ' By default all available migrations are applied.'
                                 ' Note that library currently have no ability rollback migrations')

        parser.add_argument('--database', '-d', nargs='?', type=str, required=False, choices=list(config.DATABASES.keys()),
                            help='ClickHouse database alias key from CLICKHOUSE_DATABASES django setting.'
                                 ' By default migrations are applied to all databases.')

    def handle(self, *args, **options) -> None:
        apps = [options['app_label']] if options['app_label'] else [app.name for app in django_apps.get_app_configs()]
        databases = [options['database']] if options['database'] else list(config.DATABASES.keys())
        kwargs = {'up_to': options['migration_number']} if options['migration_number'] else {}

        self.stdout.write(self.style.MIGRATE_HEADING(
            "Applying ClickHouse migrations for apps %s in databases %s" % (json.dumps(apps), json.dumps(databases))))

        any_migrations_applied = False
        for app_label in apps:
            for db_alias in databases:
                res = migrate_app(app_label, db_alias, verbosity=options['verbosity'], **kwargs)
                any_migrations_applied = any_migrations_applied or res

        if not any_migrations_applied:
            self.stdout.write("No migrations to apply")



"""
This file defines base abstract models to inherit from
"""
import datetime
import logging
from collections import defaultdict
from copy import deepcopy
from itertools import chain
from typing import List, Tuple, Iterable, Set, Any, Optional

from django.db.models import Model as DjangoModel, QuerySet as DjangoQuerySet
from django.utils.timezone import now
from infi.clickhouse_orm.engines import CollapsingMergeTree
from infi.clickhouse_orm.models import Model as InfiModel, ModelBase as InfiModelBase
from statsd.defaults.django import statsd

from .compatibility import namedtuple
from .configuration import config
from .database import connections, Database
from .exceptions import RedisLockTimeoutError
from .models import ClickHouseSyncModel
from .query import QuerySet
from .serializers import Django2ClickHouseModelSerializer
from .utils import lazy_class_import, exec_multi_arg_func

logger = logging.getLogger('django-clickhouse')


class ClickHouseModelMeta(InfiModelBase):
    def __new__(cls, *args, **kwargs):
        res = super().__new__(cls, *args, **kwargs)  # type: ClickHouseModel

        if res.django_model and not issubclass(res.django_model, ClickHouseSyncModel):
            raise TypeError('django_model must be ClickHouseSyncModel subclass')

        if res.django_model and res.sync_enabled:
            res.django_model.register_clickhouse_sync_model(res)

        res.objects = QuerySet(res)

        return res


class ClickHouseModel(InfiModel, metaclass=ClickHouseModelMeta):
    """
    Base model for all other models
    """
    django_model = None
    django_model_serializer = Django2ClickHouseModelSerializer

    # Servers, model is replicated to.
    # Router takes random database to read or write from.
    read_db_aliases = (config.DEFAULT_DB_ALIAS,)
    write_db_aliases = (config.DEFAULT_DB_ALIAS,)

    # Databases to perform replicated migration queries, such as ALTER TABLE.
    # Migration is applied to random database from the list.
    migrate_replicated_db_aliases = (config.DEFAULT_DB_ALIAS,)

    # Databases to perform non-replicated migrations (CREATE TABLE, DROP TABLE).
    # Migration is applied to all databases from the list.
    migrate_non_replicated_db_aliases = (config.DEFAULT_DB_ALIAS,)

    sync_enabled = False
    sync_batch_size = None
    sync_storage = None
    sync_delay = None
    sync_lock_timeout = None

    # This flag gives ability to disable to_db_string while inserting data, if it is already formatted
    sync_formatted_tuples = False

    # This attribute is initialized in metaclass, as it must get model class as a parameter
    objects = None  # type: QuerySet

    @classmethod
    def get_tuple_class(cls, field_names=None, defaults=None):
        field_names = field_names or cls.fields(writable=False).keys()

        if defaults:
            defaults_new = deepcopy(cls._defaults)
            defaults_new.update(defaults)
        else:
            defaults_new = cls._defaults

        # defaults should be rightmost arguments
        required_field_names = tuple(name for name in field_names if name not in defaults_new)

        default_field_names, default_values = zip(*sorted(defaults_new.items(), key=lambda t: t[0]))

        # Strange, but sometimes the columns are in different order...
        field_names = tuple(sorted(required_field_names)) + default_field_names

        return namedtuple("%sTuple" % cls.__name__, field_names, defaults=default_values)

    @classmethod
    def objects_in(cls, database: Database) -> QuerySet:
        return QuerySet(cls, database)

    @classmethod
    def get_database_alias(cls, for_write: bool = False) -> str:
        """
        Gets database alias for read or write purposes
        :param for_write: Boolean flag if database is neede for read or for write
        :return: Database alias to use
        """
        db_router = lazy_class_import(config.DATABASE_ROUTER)()
        if for_write:
            return db_router.db_for_write(cls)
        else:
            return db_router.db_for_read(cls)

    @classmethod
    def get_database(cls, for_write: bool = False) -> Database:
        """
        Gets database alias for read or write purposes
        :param for_write: Boolean flag if database is neede for read or for write
        :return: infi.clickhouse_orm.Database instance
        """
        db_alias = cls.get_database_alias(for_write=for_write)
        return connections[db_alias]

    @classmethod
    def get_django_model_serializer(cls, writable: bool = False, defaults: Optional[dict] = None
                                    ) -> Django2ClickHouseModelSerializer:
        serializer_cls = lazy_class_import(cls.django_model_serializer)
        return serializer_cls(cls, writable=writable, defaults=defaults)

    @classmethod
    def get_sync_batch_size(cls):
        return cls.sync_batch_size or config.SYNC_BATCH_SIZE

    @classmethod
    def get_storage(cls):
        return lazy_class_import(cls.sync_storage or config.SYNC_STORAGE)()

    @classmethod
    def get_sync_delay(cls):
        return cls.sync_delay or config.SYNC_DELAY

    @classmethod
    def get_lock_timeout(cls):
        return cls.sync_lock_timeout or cls.get_sync_delay() * 10

    @classmethod
    def get_import_key(cls):
        return cls.__name__

    def _prepare_val_for_eq(self, field_name, field, val):
        if isinstance(val, datetime.datetime):
            return val.replace(microsecond=0)

        # Sign column for collapsing should be ignored
        if isinstance(self.engine, CollapsingMergeTree) and field_name == self.engine.sign_col:
            return True

        return val

    def __eq__(self, other):
        if other.__class__ != self.__class__:
            return False

        for name, f in self.fields().items():
            val_1 = getattr(self, name, None)
            val_2 = getattr(other, name, None)
            if self._prepare_val_for_eq(name, f, val_1) != self._prepare_val_for_eq(name, f, val_2):
                return False

        return True

    @classmethod
    def get_sync_query_set(cls, using: str, pk_set: Set[Any]) -> DjangoQuerySet:
        """
        Forms django queryset to fetch for sync
        :param using: Database to fetch from
        :param pk_set: A set of primary keys to fetch
        :return: QuerySet
        """
        return cls.django_model.objects.filter(pk__in=pk_set).using(using)

    @classmethod
    def get_sync_objects(cls, operations: List[Tuple[str, str]]) -> List[DjangoModel]:
        """
        Returns objects from main database to sync
        :param operations: A list of operations to perform
        :return: A list of django_model instances
        """
        if not operations:
            return []

        pk_by_db = defaultdict(set)
        for op, pk_str in operations:
            using, pk = pk_str.split('.')
            pk_by_db[using].add(pk)

        # Selecting data from multiple databases should work faster in parallel, if connections are independent.
        objs = exec_multi_arg_func(
            lambda db_alias: list(cls.get_sync_query_set(db_alias, pk_by_db[db_alias])),
            pk_by_db.keys()
        )
        return list(chain(*objs))

    @classmethod
    def get_insert_batch(cls, import_objects: Iterable[DjangoModel]) -> List['ClickHouseModel']:
        """
        Formats django model objects to batch of ClickHouse objects
        :param import_objects: DjangoModel objects to import
        :return: ClickHouseModel objects to import
        """
        return cls.engine.get_insert_batch(cls, import_objects)

    @classmethod
    def insert_batch(cls, batch):
        """
        Inserts batch into database
        :param batch: Batch of tuples to insert
        :return:
        """
        if batch:
            cls.get_database(for_write=True).insert_tuples(cls, batch, formatted=cls.sync_formatted_tuples)

    @classmethod
    def sync_batch_from_storage(cls):
        """
        Gets one batch from storage and syncs it.
        :return:
        """
        import_key = cls.get_import_key()
        storage = cls.get_storage()
        statsd_key = "%s.sync.%s.{0}" % (config.STATSD_PREFIX, import_key)

        try:
            with statsd.timer(statsd_key.format('total')):
                with statsd.timer(statsd_key.format('steps.pre_sync')):
                    storage.pre_sync(import_key, lock_timeout=cls.get_lock_timeout())

                with statsd.timer(statsd_key.format('steps.get_operations')):
                    operations = storage.get_operations(import_key, cls.get_sync_batch_size())

                statsd.incr(statsd_key.format('operations'), len(operations))
                logger.debug('django-clickhouse: got %d operations from storage (key: %s)'
                             % (len(operations), import_key))

                if operations:
                    with statsd.timer(statsd_key.format('steps.get_sync_objects')):
                        import_objects = cls.get_sync_objects(operations)
                else:
                    import_objects = []

                statsd.incr(statsd_key.format('import_objects'), len(import_objects))
                logger.debug('django-clickhouse: got %d objects to import from database (key: %s)'
                             % (len(import_objects), import_key))

                if import_objects:
                    with statsd.timer(statsd_key.format('steps.get_insert_batch')):
                        # NOTE I don't use generator pattern here, as it move all time into insert.
                        #  That makes hard to understand where real problem is in monitoring
                        batch = tuple(cls.get_insert_batch(import_objects))

                    logger.debug('django-clickhouse: formed %d ClickHouse objects to insert (key: %s)'
                                 % (len(batch), import_key))

                    with statsd.timer(statsd_key.format('steps.insert')):
                        cls.insert_batch(batch)

                with statsd.timer(statsd_key.format('steps.post_sync')):
                    storage.post_sync(import_key)
        except RedisLockTimeoutError:
            pass  # skip this sync round if lock is acquired by another thread
        except Exception as ex:
            with statsd.timer(statsd_key.format('steps.post_sync')):
                storage.post_sync_failed(import_key)
            raise ex

    @classmethod
    def need_sync(cls) -> bool:
        """
        Checks if this model needs synchronization: sync is enabled and delay has passed
        :return: Boolean
        """
        if not cls.sync_enabled:
            logger.debug('django-clickhouse: need_sync returned False for class %s as sync is disabled' % cls.__name__)
            return False

        last_sync_time = cls.get_storage().get_last_sync_time(cls.get_import_key())

        if last_sync_time is None:
            logger.debug('django-clickhouse: need_sync returned True for class %s as no last sync found' % cls.__name__)
            return True

        res = (datetime.datetime.now() - last_sync_time).total_seconds() >= cls.get_sync_delay()
        logger.debug('django-clickhouse: need_sync returned %s for class %s as no last sync found'
                     ' (now: %s, last: %s, delay: %d)'
                     % (res, cls.__name__, now().isoformat(), last_sync_time.isoformat(),
                        cls.get_sync_delay()))

        return res


class ClickHouseMultiModel(ClickHouseModel):
    """
    This model syncs one django model with multiple ClickHouse sub-models
    """
    sub_models = []

    @classmethod
    def sync_batch_from_storage(cls):
        """
        Gets one batch from storage and syncs it.
        :return:
        """
        import_key = cls.get_import_key()
        storage = cls.get_storage()
        statsd_key = "%s.sync.%s.{0}" % (config.STATSD_PREFIX, import_key)

        try:
            with statsd.timer(statsd_key.format('total')):
                with statsd.timer(statsd_key.format('steps.pre_sync')):
                    storage.pre_sync(import_key, lock_timeout=cls.get_lock_timeout())

                with statsd.timer(statsd_key.format('steps.get_operations')):
                    operations = storage.get_operations(import_key, cls.get_sync_batch_size())

                statsd.incr(statsd_key.format('operations'), len(operations))
                logger.debug('django-clickhouse: got %d operations from storage (key: %s)'
                             % (len(operations), import_key))

                if operations:
                    with statsd.timer(statsd_key.format('steps.get_sync_objects')):
                        import_objects = cls.get_sync_objects(operations)
                else:
                    import_objects = []

                statsd.incr(statsd_key.format('import_objects'), len(import_objects))
                logger.debug('django-clickhouse: got %d objects to import from database (key: %s)'
                             % (len(import_objects), import_key))

                if import_objects:
                    batches = {}
                    with statsd.timer(statsd_key.format('steps.get_insert_batch')):
                        def _sub_model_func(model_cls):
                            model_statsd_key = "%s.sync.%s.{0}" % (config.STATSD_PREFIX, model_cls.__name__)
                            with statsd.timer(model_statsd_key.format('steps.get_insert_batch')):
                                # NOTE I don't use generator pattern here, as it move all time into insert.
                                # That makes hard to understand where real problem is in monitoring
                                batch = tuple(model_cls.get_insert_batch(import_objects))

                            logger.debug('django-clickhouse: formed %d ClickHouse objects to insert'
                                         ' (model_cls: %s, key: %s)' % (len(batch), model_cls.__name__, import_key))
                            return model_cls, batch

                        res = exec_multi_arg_func(_sub_model_func, cls.sub_models, threads_count=len(cls.sub_models))
                        batches = dict(res)

                    with statsd.timer(statsd_key.format('steps.insert')):
                        def _sub_model_func(model_cls):
                            model_statsd_key = "%s.sync.%s.{0}" % (config.STATSD_PREFIX, model_cls.__name__)
                            with statsd.timer(model_statsd_key.format('steps.insert')):
                                model_cls.insert_batch(batches[model_cls])

                        exec_multi_arg_func(_sub_model_func, cls.sub_models, threads_count=len(cls.sub_models))

                with statsd.timer(statsd_key.format('steps.post_sync')):
                    storage.post_sync(import_key)

        except RedisLockTimeoutError:
            pass  # skip this sync round if lock is acquired by another thread
        except Exception as ex:
            with statsd.timer(statsd_key.format('steps.post_sync')):
                storage.post_sync_failed(import_key)
            raise ex



from datetime import date

from django.test import TestCase

from django_clickhouse.migrations import migrate_app
from tests.clickhouse_models import ClickHouseTestModel


class CollapsingMergeTreeTest(TestCase):
    maxDiff = None

    def setUp(self):
        self.db = ClickHouseTestModel.get_database(for_write=True)
        self.db.drop_database()
        self.db.create_database()
        migrate_app('tests', 'default')

    def test_insert_tuples(self):
        tuple_class = ClickHouseTestModel.get_tuple_class()
        data = [
            tuple_class(id=i, created_date=date.today(), value=i, str_field=str(i))
            for i in range(10)
        ]
        self.db.insert_tuples(ClickHouseTestModel, data)

        qs = ClickHouseTestModel.objects.order_by('id').all()
        self.assertListEqual([{
            'id': i,
            'created_date': date.today(),
            'value': i,
            'str_field': str(i)
        } for i in range(10)], [item.to_dict() for item in qs])

    def test_insert_tuples_defaults(self):
        tuple_class = ClickHouseTestModel.get_tuple_class(defaults={'created_date': date.today()})
        data = [
            tuple_class(id=i, str_field=str(i))
            for i in range(10)
        ]
        self.db.insert_tuples(ClickHouseTestModel, data)

        qs = ClickHouseTestModel.objects.order_by('id').all()
        self.assertListEqual([{
            'id': i,
            'created_date': date.today(),
            'value': 100500,
            'str_field': str(i)
        } for i in range(10)], [item.to_dict() for item in qs])

    def test_insert_tuples_batch_size(self):
        tuple_class = ClickHouseTestModel.get_tuple_class()
        data = [
            tuple_class(id=i, created_date=date.today(), value=i, str_field=str(i))
            for i in range(10)
        ]
        self.db.insert_tuples(ClickHouseTestModel, data, batch_size=2)

        qs = ClickHouseTestModel.objects.order_by('id').all()
        self.assertListEqual([{
            'id': i,
            'created_date': date.today(),
            'value': i,
            'str_field': str(i)
        } for i in range(10)], [item.to_dict() for item in qs])

    def test_insert_tuples_special_characters(self):
        tuple_class = ClickHouseTestModel.get_tuple_class()
        data = [tuple_class(id=1, created_date=date.today(), value=1, str_field='\t')]
        self.db.insert_tuples(ClickHouseTestModel, data)

        item = ClickHouseTestModel.objects.filter(id=1)[0]
        self.assertEqual('\t', item.str_field)

        data = [tuple_class(id=2, created_date=date.today(), value=2, str_field='\n')]
        self.db.insert_tuples(ClickHouseTestModel, data)

        item = ClickHouseTestModel.objects.filter(id=2)[0]
        self.assertEqual('\n', item.str_field)

    def test_select_tuples(self):
        ClickHouseTestModel.objects.bulk_create([
            ClickHouseTestModel(id=i, created_date=date.today(), value=i, str_field=str(i))
            for i in range(10)
        ])

        res = self.db.select_tuples('SELECT * FROM $table ORDER BY id', ClickHouseTestModel)
        tuple_class = ClickHouseTestModel.get_tuple_class()
        self.assertListEqual([
            tuple_class(id=i, created_date=date.today(), value=i, str_field=str(i))
            for i in range(10)
        ], list(res))



import datetime

from django.test import TestCase

from django_clickhouse.exceptions import RedisLockTimeoutError
from django_clickhouse.storages import RedisStorage
from tests.clickhouse_models import ClickHouseTestModel, ClickHouseCollapseTestModel


class StorageTest(TestCase):
    storage = RedisStorage()

    def setUp(self):
        self.storage.flush()

    def tearDown(self):
        self.storage.flush()

    def test_operation_pks(self):
        self.storage.register_operations_wrapped('test', 'insert', 100500)
        self.storage.register_operations_wrapped('test', 'insert', 100501)
        self.storage.register_operations_wrapped('test', 'insert', 100502)
        self.assertListEqual([
            ('insert', '100500'),
            ('insert', '100501'),
            ('insert', '100502'),
        ], self.storage.get_operations('test', 10))

    def test_operation_types(self):
        self.storage.register_operations_wrapped('test', 'insert', 100500)
        self.storage.register_operations_wrapped('test', 'update', 100500)
        self.storage.register_operations_wrapped('test', 'delete', 100500)
        self.assertListEqual([
            ('insert', '100500'),
            ('update', '100500'),
            ('delete', '100500'),
        ], self.storage.get_operations('test', 10))

    def test_operation_import_keys(self):
        self.storage.register_operations_wrapped('test1', 'insert', 100500)
        self.storage.register_operations_wrapped('test2', 'insert', 100500)
        self.storage.register_operations_wrapped('test2', 'insert', 100501)
        self.assertListEqual([
            ('insert', '100500')
        ], self.storage.get_operations('test1', 10))
        self.assertListEqual([
            ('insert', '100500'),
            ('insert', '100501'),
        ], self.storage.get_operations('test2', 10))

    def test_post_sync(self):
        self.storage.pre_sync('test')
        self.storage.register_operations_wrapped('test', 'insert', 100500)
        self.storage.register_operations_wrapped('test', 'insert', 100501)
        self.storage.get_operations('test', 10)
        self.storage.register_operations_wrapped('test', 'insert', 100502)

        self.storage.post_sync('test')
        self.assertListEqual([
            ('insert', '100502')
        ], self.storage.get_operations('test', 10))

    def test_last_sync(self):
        dt = datetime.datetime.now()
        self.storage.set_last_sync_time('test', dt)
        self.assertEqual(dt, self.storage.get_last_sync_time('test'))

    def test_operations_count(self):
        self.storage.register_operations_wrapped('test', 'insert', 100500)
        self.storage.register_operations_wrapped('test', 'insert', 100501)
        self.assertEqual(2, self.storage.operations_count('test'))
        self.storage.register_operations_wrapped('test', 'insert', 100502)
        self.assertEqual(3, self.storage.operations_count('test'))

    def test_locks(self):
        # Test that multiple can acquire locks in parallel
        # And single model can't
        lock = self.storage.get_lock(ClickHouseTestModel.get_import_key())
        lock.acquire()
        with self.assertRaises(RedisLockTimeoutError):
            lock.acquire()

        lock_2 = self.storage.get_lock(ClickHouseCollapseTestModel.get_import_key())
        lock_2.acquire()



from typing import List, Dict, Any
from unittest import mock

from django.conf import settings
from django.test import TestCase, override_settings

from django_clickhouse.configuration import config
from django_clickhouse.database import connections
from django_clickhouse.management.commands.clickhouse_migrate import Command
from django_clickhouse.migrations import MigrationHistory, migrate_app
from django_clickhouse.routers import DefaultRouter
from tests.clickhouse_models import ClickHouseTestModel


class NoMigrateRouter(DefaultRouter):
    def allow_migrate(self, db_alias, app_label, operation, **hints):
        return False


def table_exists(db, model_class):
    res = db.select(
        "SELECT * FROM system.tables WHERE `database`='%s' AND `name`='%s'"
        % (db.db_name, model_class.table_name())
    )
    res = list(res)
    return bool(res)


@override_settings(CLICKHOUSE_MIGRATE_WITH_DEFAULT_DB=False)
class MigrateAppTest(TestCase):
    def setUp(self):
        self.db = connections['default']

        # Clean all database data
        self.db.drop_database()
        self.db.db_exists = False
        self.db.create_database()

    def test_migrate_app(self):
        migrate_app('tests', 'default')
        self.assertTrue(table_exists(self.db, ClickHouseTestModel))

        self.assertEqual(1, self.db.count(MigrationHistory))

        # Migrations are already applied no actions should be done
        migrate_app('tests', 'default')
        self.assertEqual(1, self.db.count(MigrationHistory))

    @override_settings(CLICKHOUSE_DATABASE_ROUTER=NoMigrateRouter)
    def test_router_not_allowed(self):
        migrate_app('tests', 'default')
        self.assertFalse(table_exists(self.db, ClickHouseTestModel))

    def test_no_migrate_connections(self):
        migrate_app('tests', 'no_migrate')
        self.assertFalse(table_exists(connections['no_migrate'], ClickHouseTestModel))

    def test_readonly_connections(self):
        migrate_app('tests', 'readonly')
        self.assertFalse(table_exists(connections['readonly'], ClickHouseTestModel))


@override_settings(CLICKHOUSE_MIGRATE_WITH_DEFAULT_DB=False)
@mock.patch('django_clickhouse.management.commands.clickhouse_migrate.migrate_app', return_value=True)
class MigrateDjangoCommandTest(TestCase):
    APP_LABELS = ('src', 'tests')

    def setUp(self) -> None:
        self.cmd = Command()

    def test_handle_all(self, migrate_app_mock):
        self.cmd.handle(verbosity=3, app_label=None, database=None, migration_number=None)

        self.assertEqual(len(config.DATABASES.keys()) * len(self.APP_LABELS), migrate_app_mock.call_count)
        for db_alias in config.DATABASES.keys():
            for app_label in self.APP_LABELS:
                migrate_app_mock.assert_any_call(app_label, db_alias, verbosity=3)

    def test_handle_app(self, migrate_app_mock):
        self.cmd.handle(verbosity=3, app_label='tests', database=None, migration_number=None)

        self.assertEqual(len(config.DATABASES.keys()), migrate_app_mock.call_count)
        for db_alias in config.DATABASES.keys():
            migrate_app_mock.assert_any_call('tests', db_alias, verbosity=3)

    def test_handle_database(self, migrate_app_mock):
        self.cmd.handle(verbosity=3, database='default', app_label=None, migration_number=None)

        self.assertEqual(len(settings.INSTALLED_APPS), migrate_app_mock.call_count)
        for app_label in self.APP_LABELS:
            migrate_app_mock.assert_any_call(app_label, 'default', verbosity=3)

    def test_handle_app_and_database(self, migrate_app_mock):
        self.cmd.handle(verbosity=3, app_label='tests', database='default', migration_number=None)

        migrate_app_mock.assert_called_with('tests', 'default', verbosity=3)

    def test_handle_migration_number(self, migrate_app_mock):
        self.cmd.handle(verbosity=3, database='default', app_label='tests', migration_number=1)

        migrate_app_mock.assert_called_with('tests', 'default', up_to=1, verbosity=3)

    def _test_parser_results(self, argv: List[str], expected: Dict[str, Any]) -> None:
        """
        Tests if parser process input correctly.
        Checks only expected parameters, ignores others.
        :param argv: List of string arguments from command line
        :param expected: Dictionary of expected results
        :return: None
        :raises AssertionError: If expected result is incorrect
        """
        parser = self.cmd.create_parser('./manage.py', 'clickhouse_migrate')

        options = parser.parse_args(argv)

        # Copied from django.core.management.base.BaseCommand.run_from_argv('...')
        cmd_options = vars(options)
        cmd_options.pop('args', ())

        self.assertDictEqual(expected, {opt: cmd_options[opt] for opt in expected.keys()})

    def test_parser(self, _):
        with self.subTest('Simple'):
            self._test_parser_results([], {
                'app_label': None,
                'database': None,
                'migration_number': None,
                'verbosity': 1
            })

        with self.subTest('App label'):
            self._test_parser_results(['tests'], {
                'app_label': 'tests',
                'database': None,
                'migration_number': None,
                'verbosity': 1
            })

        with self.subTest('App label and migration number'):
            self._test_parser_results(['tests', '123'], {
                'app_label': 'tests',
                'database': None,
                'migration_number': 123,
                'verbosity': 1
            })

        with self.subTest('Database'):
            self._test_parser_results(['--database', 'default'], {
                'app_label': None,
                'database': 'default',
                'migration_number': None,
                'verbosity': 1
            })

        with self.subTest('Verbosity'):
            self._test_parser_results(['--verbosity', '2'], {
                'app_label': None,
                'database': None,
                'migration_number': None,
                'verbosity': 2
            })



from unittest import TestCase

import datetime

from django_clickhouse.database import connections
from django_clickhouse.migrations import migrate_app
from django_clickhouse.query import QuerySet
from tests.clickhouse_models import ClickHouseTestModel


class TestQuerySet(TestCase):
    def _recreate_db(self, db_alias):
        db = connections[db_alias]
        db.drop_database()
        db.db_exists = False
        db.create_database()

        migrate_app('tests', db_alias)
        return db

    def setUp(self):
        self.db = self._recreate_db('default')
        self._recreate_db('secondary')

    def test_all(self):
        self.db.insert([ClickHouseTestModel(id=i, created_date=datetime.date.today(), value=i) for i in range(1, 4)])
        qs = ClickHouseTestModel.objects.all()

        self.assertIsInstance(qs, QuerySet)
        self.assertEqual(3, qs.count())

    def test_create(self):
        ClickHouseTestModel.objects.create(id=1, created_date=datetime.date.today(), value=2)
        res = list(self.db.select('SELECT * FROM $table', model_class=ClickHouseTestModel))
        self.assertEqual(1, len(res))

        self.assertEqual(1, res[0].id)
        self.assertEqual(datetime.date.today(), res[0].created_date)
        self.assertEqual(2, res[0].value)

    def test_bulk_create(self):
        ClickHouseTestModel.objects.bulk_create([
            ClickHouseTestModel(id=i, created_date=datetime.date.today(), value=i) for i in range(1, 4)
        ])
        res = list(self.db.select('SELECT * FROM $table ORDER BY id', model_class=ClickHouseTestModel))
        self.assertEqual(3, len(res))

        for i in range(0, 3):
            self.assertEqual(i + 1, res[i].id)
            self.assertEqual(datetime.date.today(), res[0].created_date)
            self.assertEqual(i + 1, res[i].value)

    def test_using(self):
        self.db.insert(
            [ClickHouseTestModel(id=i, created_date=datetime.date.today(), value=i) for i in range(1, 4)]
        )
        connections['secondary'].insert([
            ClickHouseTestModel(id=i, created_date=datetime.date.today(), value=i) for i in range(10, 12)
        ])

        self.assertEqual(3, ClickHouseTestModel.objects.count())
        self.assertEqual(3, ClickHouseTestModel.objects_in(self.db).count())
        self.assertEqual(2, ClickHouseTestModel.objects_in(self.db).using('secondary').count())

        self.assertEqual(2, ClickHouseTestModel.objects.using('secondary').count())
        self.assertEqual(3, ClickHouseTestModel.objects.using('default').count())



import datetime
from queue import Queue, Empty
from threading import Thread

import os
from itertools import chain
from typing import Union, Any, Optional, TypeVar, Set, Dict, Iterable, Tuple, Iterator, Callable, List

import pytz
from importlib import import_module
from importlib.util import find_spec
from django.db.models import Model as DjangoModel

from .database import connections


T = TypeVar('T')


def get_tz_offset(db_alias: Optional[str] = None) -> int:
    """
    Returns ClickHouse server timezone offset in minutes
    :param db_alias: The database alias used
    :return: Integer
    """
    db = connections[db_alias]
    return int(db.server_timezone.utcoffset(datetime.datetime.utcnow()).total_seconds() / 60)


def format_datetime(dt: Union[datetime.date, datetime.datetime], timezone_offset: int = 0, day_end: bool = False,
                    db_alias: Optional[str] = None) -> str:
    """
    Formats datetime and date objects to format that can be used in WHERE conditions of query
    :param dt: datetime.datetime or datetime.date object
    :param timezone_offset: timezone offset (minutes)
    :param day_end: If datetime.date is given and flag is set, returns day end time, not day start.
    :param db_alias: The database alias used
    :return: A string representing datetime
    """
    assert isinstance(dt, (datetime.datetime, datetime.date)), "dt must be datetime.datetime instance"
    assert type(timezone_offset) is int, "timezone_offset must be integer"

    # datetime.datetime inherits datetime.date. So I can't just make isinstance(dt, datetime.date)
    if not isinstance(dt, datetime.datetime):
        t = datetime.time.max if day_end else datetime.time.min
        dt = datetime.datetime.combine(dt, t)

    # Convert datetime to UTC, if it has timezone
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        dt = pytz.utc.localize(dt)
    else:
        dt = dt.astimezone(pytz.utc)

    # Dates in ClickHouse are parsed in server local timezone. So I need to add server timezone
    server_dt = dt - datetime.timedelta(minutes=timezone_offset - get_tz_offset(db_alias))

    return server_dt.strftime("%Y-%m-%d %H:%M:%S")


def module_exists(module_name: str) -> bool:
    """
    Checks if module exists
    :param module_name: Dot-separated module name
    :return: Boolean
    """
    # Python 3.4+
    spam_spec = find_spec(module_name)
    return spam_spec is not None


def lazy_class_import(obj: Union[str, Any]) -> Any:
    """
    If string is given, imports object by given module path.
    Otherwise returns the object
    :param obj: A string class path or object to return
    :return: Imported object
    """
    if isinstance(obj, str):
        module_name, obj_name = obj.rsplit('.', 1)
        module = import_module(module_name)

        try:
            return getattr(module, obj_name)
        except AttributeError:
            raise ImportError('Invalid import path `%s`' % obj)
    else:
        return obj


def get_subclasses(cls: T, recursive: bool = False) -> Set[T]:
    """
    Gets all subclasses of given class
    Attention!!! Classes would be found only if they were imported before using this function
    :param cls: Class to get subcalsses
    :param recursive: If flag is set, returns subclasses of subclasses and so on too
    :return: A list of subclasses
    """
    subclasses = set(cls.__subclasses__())

    if recursive:
        for subcls in subclasses.copy():
            subclasses.update(get_subclasses(subcls, recursive=True))

    return subclasses


def model_to_dict(instance: DjangoModel, fields:  Optional[Iterable[str]] = None,
                  exclude_fields:  Optional[Iterable[str]] = None) -> Dict[str, Any]:
    """
    Standard model_to_dict ignores some fields if they have invalid naming
    :param instance: Object to convert to dictionary
    :param fields: Field list to extract from instance
    :param exclude_fields: Filed list to exclude from extraction
    :return: Serialized dictionary
    """
    data = {}

    if not fields:
        opts = instance._meta
        fields = {f.name for f in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many)}

    for name in set(fields) - set(exclude_fields or set()):
        val = getattr(instance, name, None)
        if val is not None:
            data[name] = val

    return data


def check_pid_exists(pid):
    """
    Check For the existence of a unix pid.
    """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def int_ranges(items: Iterable[int]) -> Iterator[Tuple[int, int]]:
    """
    Finds continuous intervals in integer iterable.
    :param items: Items to search in
    :return: Iterator over Tuple[start, end]
    """
    interval_start = None
    prev_item = None
    for item in sorted(items):
        if prev_item is None:
            interval_start = prev_item = item
        elif prev_item + 1 == item:
            prev_item = item
        else:
            interval = interval_start, prev_item
            interval_start = prev_item = item
            yield interval

    if interval_start is None:
        return
    else:
        yield interval_start, prev_item


class ExceptionThread(Thread):
    """
    Thread objects, which catches thread exceptions and raises them in main thread
    """
    def __init__(self, *args, **kwargs):
        super(ExceptionThread, self).__init__(*args, **kwargs)
        self.exc = None

    def _close_django_db_connections(self):
        """
        In Django every thread has its own database connection pool.
        But django does not close them automatically in child threads.
        As a result, this can cause database connection leaking.
        Here we close connections manually when thread execution is finished.
        """
        try:
            from django.db import connections as db_connections
        except (ModuleNotFoundError, ImportError):
            db_connections = None

        if db_connections:
            db_connections.close_all()

    def run(self):
        try:
            return super(ExceptionThread, self).run()
        except Exception as e:
            self.exc = e
        finally:
            self._close_django_db_connections()

    def join(self, timeout=None):
        super(ExceptionThread, self).join(timeout=timeout)
        if self.exc:
            raise self.exc


def exec_in_parallel(func: Callable, args_queue: Queue, threads_count: Optional[int] = None) -> List[Any]:
    """
    Executes func in multiple threads in parallel
    Functions are expected to be thread safe. If it needs some locks, func must provide them.
    :param func: Function to execute in thread
    :param args_queue: A queue with arguments for separate function call. Each element is tuple of (args, kwargs)
    :param threads_count: Maximum number of parallel threads tho run
    :return: A list of results. Order of results is not guaranteed. Element types depends func return type.
    """
    results = []

    # If thread_count is not given, we execute all tasks in parallel.
    # If queue has less elements than threads_count, take queue size.
    threads_count = min(args_queue.qsize(), threads_count) if threads_count else args_queue.qsize()

    def _worker():
        """
        Thread worker, gets next arguments from queue and processes them.
        Results are put into results array using thread safe lock
        :return: None
        """
        finished = False
        while not finished:
            try:
                # Get arguments
                args, kwargs = args_queue.get_nowait()

                # Execute function
                local_res = func(*args, **kwargs)

                # Write result. appending a list is thread safe operation according to:
                # http://effbot.org/pyfaq/what-kinds-of-global-value-mutation-are-thread-safe.htm
                results.append(local_res)

                # Mark task as complete
                args_queue.task_done()
            except Empty:
                # No data in queue, finish worker thread
                finished = True

    # Run threads
    threads = []
    for index in range(threads_count):
        t = ExceptionThread(target=_worker)
        threads.append(t)
        t.start()

    # Wait for threads to finish
    for t in threads:
        t.join()

    return results


def exec_multi_arg_func(func: Callable, split_args: Iterable[Any], *args, threads_count: Optional[int] = None,
                        **kwargs) -> List[Any]:
    """
    Executes function in parallel threads. Thread functions (func) receive one of split_args as first argument
    Another arguments passed to functions - args and kwargs
    If len(split_args) <= 0, separate threads are not run, main thread is used.
    :param func: Function to execute. Must accept split_arg as first parameter
    :param split_args: A list of arguments to split threads by
    :param threads_count: Maximum number of threads to run in parallel
    :return: A list of execution results. Order of execution is not guaranteed.
    """
    split_args = list(split_args)
    if len(split_args) == 0:
        return []
    elif len(split_args) == 1:
        return [func(split_args[0], *args, **kwargs)]
    else:
        q = Queue()
        for s in split_args:
            q.put(([s] + list(args), kwargs))

        return exec_in_parallel(func, q, threads_count=threads_count)


class SingletonMeta(type):
    """
    Realises singleton pattern
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



"""
This file contains base django model to be synced with ClickHouse.
It saves all operations to storage in order to write them to ClickHouse later.
"""

from typing import Optional, Any, Type, Set

from django.db import transaction
from django.db.models import QuerySet as DjangoQuerySet, Model as DjangoModel, Manager as DjangoManager
from django.db.models.manager import BaseManager
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from statsd.defaults.django import statsd

from .compatibility import update_returning_pk
from .configuration import config
from .storages import Storage
from .utils import lazy_class_import

try:
    from django_pg_returning.manager import UpdateReturningMixin
except ImportError:
    class UpdateReturningMixin:
        fake = True


try:
    from django_pg_bulk_update.manager import BulkUpdateManagerMixin
except ImportError:
    class BulkUpdateManagerMixin:
        fake = True


class ClickHouseSyncRegisterMixin:
    def _register_ops(self, operation, result, as_int: bool = False):
        pk_name = self.model._meta.pk.name
        pk_list = [getattr(item, pk_name) if isinstance(item, DjangoModel) else item for item in result]
        self.model.register_clickhouse_operations(operation, *pk_list, using=self.db)


class ClickHouseSyncUpdateReturningQuerySetMixin(ClickHouseSyncRegisterMixin, UpdateReturningMixin):
    """
    This mixin adopts methods of django-pg-returning library
    """

    def update_returning(self, **updates):
        result = super().update_returning(**updates)
        self._register_ops('update', result)
        return result

    def delete_returning(self):
        result = super().delete_returning()
        self._register_ops('delete', result)
        return result


class ClickHouseSyncBulkUpdateQuerySetMixin(ClickHouseSyncRegisterMixin, BulkUpdateManagerMixin):
    """
    This mixin adopts methods of django-pg-bulk-update library
    """

    def _update_returning_param(self, returning):
        pk_name = self.model._meta.pk.name
        if returning is None:
            returning = pk_name
        elif isinstance(returning, str):
            returning = [pk_name, returning] if returning != '*' else '*'
        else:
            returning = list(returning) + [pk_name]

        return returning

    def _decorate_method(self, name: str, operation: str, args, kwargs):
        if not hasattr(super(), name):
            raise AttributeError("QuerySet has no attribute %s. Is django-pg-bulk-update library installed?" % name)

        func = getattr(super(), name)
        original_returning = kwargs.pop('returning', None)
        kwargs['returning'] = self._update_returning_param(original_returning)
        result = func(*args, **kwargs)
        self._register_ops(operation, result)
        return result.count() if original_returning is None else result

    def pg_bulk_update(self, *args, **kwargs):
        return self._decorate_method('pg_bulk_update', 'update', args, kwargs)

    def pg_bulk_update_or_create(self, *args, **kwargs):
        return self._decorate_method('pg_bulk_update_or_create', 'update', args, kwargs)

    def pg_bulk_create(self, *args, **kwargs):
        return self._decorate_method('pg_bulk_create', 'insert', args, kwargs)


class ClickHouseSyncQuerySetMixin(ClickHouseSyncRegisterMixin):
    def update(self, **kwargs):
        pks = update_returning_pk(self, kwargs)
        self._register_ops('update', pks)
        return len(pks)

    def bulk_create(self, objs, batch_size=None):
        objs = super().bulk_create(objs, batch_size=batch_size)
        self._register_ops('insert', objs)
        return objs

    def bulk_update(self, objs, *args, **kwargs):
        objs = list(objs)

        # No need to register anything, if there are no objects.
        # If objects are not models, django-pg-bulk-update method is called and pg_bulk_update will register items
        if len(objs) == 0 or not isinstance(objs[0], DjangoModel):
            return super().bulk_update(objs, *args, **kwargs)

        # native django bulk_update requires each object to have a primary key
        res = super().bulk_update(objs, *args, **kwargs)
        self._register_ops('update', objs)
        return res


# I add library dependant mixins to base classes only if libraries are installed
qs_bases = [ClickHouseSyncQuerySetMixin]

if not getattr(UpdateReturningMixin, 'fake', False):
    qs_bases.append(ClickHouseSyncUpdateReturningQuerySetMixin)

if not getattr(BulkUpdateManagerMixin, 'fake', False):
    qs_bases.append(ClickHouseSyncBulkUpdateQuerySetMixin)


# QuerySet must be the last one, so it can be redeclared in mixins
qs_bases.append(DjangoQuerySet)
ClickHouseSyncQuerySet = type('ClickHouseSyncModelQuerySet', tuple(qs_bases), {})


class ClickHouseSyncManager(BaseManager.from_queryset(ClickHouseSyncQuerySet), DjangoManager):
    pass


class ClickHouseSyncModel(DjangoModel):
    """
    Base model for syncing data. Each django model synced with data must inherit this
    """
    objects = ClickHouseSyncManager()

    class Meta:
        abstract = True

    @classmethod
    def get_clickhouse_storage(cls) -> Storage:
        """
        Returns Storage instance to save clickhouse sync data to
        :return:
        """
        storage_cls = lazy_class_import(config.SYNC_STORAGE)
        return storage_cls()

    @classmethod
    def register_clickhouse_sync_model(cls, model_cls: Type['ClickHouseModel']) -> None:  # noqa: F821
        """
        Registers ClickHouse model to listen to this model updates
        :param model_cls: Model class to register
        :return: None
        """
        if not hasattr(cls, '_clickhouse_sync_models'):
            cls._clickhouse_sync_models = set()

        cls._clickhouse_sync_models.add(model_cls)

    @classmethod
    def get_clickhouse_sync_models(cls) -> Set['ClickHouseModel']:  # noqa: F821
        """
        Returns all clickhouse models, listening to this class
        :return: A set of model classes to sync
        """
        return getattr(cls, '_clickhouse_sync_models', set())

    @classmethod
    def register_clickhouse_operations(cls, operation: str, *model_pks: Any, using: Optional[str] = None) -> None:
        """
        Registers model operation in storage
        :param operation: Operation type - one of [insert, update, delete)
        :param model_pks: Elements to import
        :param using: Database alias registered instances are from
        :return: None
        """
        model_pks = ['%s.%s' % (using or config.DEFAULT_DB_ALIAS, pk) for pk in model_pks]

        def _on_commit():
            for model_cls in cls.get_clickhouse_sync_models():
                if model_cls.django_model == cls:
                    storage.register_operations_wrapped(model_cls.get_import_key(), operation, *model_pks)

        if len(model_pks) > 0:
            storage = cls.get_clickhouse_storage()
            transaction.on_commit(_on_commit, using=using)

    def post_save(self, created: bool, using: Optional[str] = None) -> None:
        self.register_clickhouse_operations('insert' if created else 'update', self.pk, using=using)

    def post_delete(self, using: Optional[str] = None) -> None:
        self.register_clickhouse_operations('delete', self.pk, using=using)


@receiver(post_save)
def post_save(sender, instance, **kwargs):
    statsd.incr('%s.sync.post_save' % config.STATSD_PREFIX, 1)
    if issubclass(sender, ClickHouseSyncModel):
        instance.post_save(kwargs.get('created', False), using=kwargs.get('using'))


@receiver(post_delete)
def post_delete(sender, instance, **kwargs):
    if issubclass(sender, ClickHouseSyncModel):
        instance.post_delete(using=kwargs.get('using'))



"""
Contains additional components for redis-py to use in RedisStorage
"""
import os
from itertools import chain
import logging

from .exceptions import RedisLockTimeoutError


logger = logging.getLogger('django-clickhouse')


class RedisLock:
    """
    Fixes issue of https://github.com/andymccurdy/redis-py/issues/621
    """

    def __init__(self, redis_client, *args, **kwargs):
        self.lock = redis_client.lock(*args, **kwargs)

    def __enter__(self):
        return self.lock.acquire()

    def __exit__(self, type, value, tb):
        self.lock.release()

    def acquire(self):
        logger.debug('django-clickhouse: acquiring lock "%s" with pid %d' % (self.lock.name, os.getpid()))
        if self.lock.acquire():
            logger.debug('django-clickhouse:  acquired lock "%s" with pid %d' % (self.lock.name, os.getpid()))
            return self
        else:
            logger.warning('django-clickhouse: timeout lock "%s" with pid %d' % (self.lock.name, os.getpid()))
            raise RedisLockTimeoutError()

    def release(self):
        logger.debug('django-clickhouse: releasing lock "%s" with pid %d' % (self.lock.name, os.getpid()))
        self.lock.release()

    def hard_release(self) -> bool:
        """
        Drops the lock, not looking if it is acquired by anyone.
        :return: Boolean - if lock has been acquired before releasing or not
        """
        logger.warning('django-clickhouse: hard releasing lock "%s" with pid %d' % (self.lock.name, os.getpid()))
        return bool(self.lock.redis.delete(self.lock.name))


def redis_zadd(redis_client, key, mapping, **kwargs):
    """
    In redis-py 3.* interface of zadd changed to mapping
    :return:
    """
    import redis
    if int(redis.__version__.split('.', 1)[0]) < 3:
        # key, score1, value1, score2, value2, ...
        items = chain(*((score, key) for key, score in mapping.items()))
    else:
        items = [mapping]

    return redis_client.zadd(key, *items, **kwargs)



"""
This file is a wrapper for django settings,
which searches for library properties and sets defaults
"""

from django.conf import settings
from typing import Any

# Prefix of all library parameters
PREFIX = getattr(settings, 'CLICKHOUSE_SETTINGS_PREFIX', 'CLICKHOUSE_')

# Default values for all library parameters
DEFAULTS = {
    'DATABASES': {},
    'SYNC_BATCH_SIZE': 10000,
    'SYNC_STORAGE': 'django_clickhouse.storages.RedisStorage',
    'SYNC_DELAY': 5,
    'REDIS_CONFIG': None,
    'MODELS_MODULE': 'clickhouse_models',
    'DATABASE_ROUTER': 'django_clickhouse.routers.DefaultRouter',
    'STATSD_PREFIX': 'clickhouse',
    'MIGRATIONS_PACKAGE': 'clickhouse_migrations',
    'MIGRATION_HISTORY_MODEL': 'django_clickhouse.migrations.MigrationHistory',
    'MIGRATE_WITH_DEFAULT_DB': True,
    'CELERY_QUEUE': 'celery',
    'DEFAULT_DB_ALIAS': 'default'
}


class Config:
    def __getattr__(self, item: str) -> Any:
        if item not in DEFAULTS:
            raise AttributeError('Unknown config parameter `%s`' % item)

        name = PREFIX + item
        return getattr(settings, name, DEFAULTS[item])


config = Config()



from typing import NamedTuple, Optional, Iterable, Type

import pytz
from django.db.models import Model as DjangoModel

from .utils import model_to_dict


class Django2ClickHouseModelSerializer:
    def __init__(self, model_cls: Type['ClickHouseModel'], fields: Optional[Iterable[str]] = None,  # noqa: F821
                 exclude_fields: Optional[Iterable[str]] = None, writable: bool = False,
                 defaults: Optional[dict] = None) -> None:
        """
        Initializes serializer
        :param model_cls: ClickHouseModel subclass to serialize to
        :param fields: Optional. A list of fields to add into result tuple
        :param exclude_fields: Fields to exclude from result tuple
        :param writable: If fields parameter is not set directly,
          this flags determines if only writable or all fields should be taken from model_cls
        :param defaults: A dictionary of field: value which are taken as default values for model_cls instances
        :return: None
        """
        self._model_cls = model_cls
        if fields is not None:
            self.serialize_fields = fields
        else:
            self.serialize_fields = model_cls.fields(writable=writable).keys()

        self.exclude_serialize_fields = exclude_fields
        self._result_class = self._model_cls.get_tuple_class(defaults=defaults)
        self._fields = self._model_cls.fields(writable=False)

    def _get_serialize_kwargs(self, obj: DjangoModel) -> dict:
        data = model_to_dict(obj, fields=self.serialize_fields, exclude_fields=self.exclude_serialize_fields)

        # Remove None values, they should be initialized as defaults
        result = {
            key: self._fields[key].to_python(value, pytz.utc)
            for key, value in data.items() if value is not None
        }

        return result

    def serialize(self, obj: DjangoModel) -> NamedTuple:
        return self._result_class(**self._get_serialize_kwargs(obj))



"""
Migrating database
"""
import datetime
from typing import Optional, Set

from django.db import DEFAULT_DB_ALIAS as DJANGO_DEFAULT_DB_ALIAS
from django.db.models.signals import post_migrate
from django.dispatch import receiver

# In order to support all operations import here
from infi.clickhouse_orm.migrations import *  # noqa F401, F403
from infi.clickhouse_orm.migrations import RunSQL as LibRunSQL, RunPython as LibRunPython

from infi.clickhouse_orm.database import ServerError, DatabaseException
from infi.clickhouse_orm.fields import StringField, DateField
from infi.clickhouse_orm.engines import MergeTree
from infi.clickhouse_orm.utils import import_submodules

from .clickhouse_models import ClickHouseModel
from .configuration import config
from .database import connections, Database
from .utils import lazy_class_import, module_exists


class Migration:
    """
    Base class for migrations
    """
    operations = []

    def apply(self, db_alias: str, database: Optional[Database] = None) -> None:
        """
        Applies migration to given database
        :param db_alias: Database alias to apply migration to
        :param database: Sometimes I want to pass db object directly for testing purposes
        :return: None
        """
        db_router = lazy_class_import(config.DATABASE_ROUTER)()
        database = database or connections[db_alias]

        for op in self.operations:
            hints = getattr(op, 'hints', {})

            if db_router.allow_migrate(db_alias, self.__module__, op, **hints):
                op.apply(database)


def migrate_app(app_label: str, db_alias: str, up_to: int = 9999, database: Optional[Database] = None,
                verbosity: int = 1) -> bool:
    """
    Migrates given django app
    :param app_label: App label to migrate
    :param db_alias: Database alias to migrate
    :param up_to: Migration number to migrate to
    :param database: Sometimes I want to pass db object directly for testing purposes
    :param verbosity: 0-4, уровень verbosity вывода
    :return: True if any migration has been applied
    """
    # Can't migrate such connection, just skip it
    if config.DATABASES[db_alias].get('readonly', False):
        if verbosity > 1:
            print('Skipping database "%s": marked as readonly' % db_alias)
        return False

    # Ignore force not migrated databases
    if not config.DATABASES[db_alias].get('migrate', True):
        if verbosity > 1:
            print('Skipping database "%s": migrations are restricted in configuration' % db_alias)
        return False

    migrations_package = "%s.%s" % (app_label, config.MIGRATIONS_PACKAGE)

    if not module_exists(migrations_package):
        if verbosity > 1:
            print('Skipping migrations for app "%s": no migration_package "%s"' % (app_label, migrations_package))
        return False

    database = database or connections[db_alias]
    migration_history_model = lazy_class_import(config.MIGRATION_HISTORY_MODEL)

    applied_migrations = migration_history_model.get_applied_migrations(db_alias, migrations_package)
    modules = import_submodules(migrations_package)
    unapplied_migrations = set(modules.keys()) - applied_migrations

    any_applied = False
    for name in sorted(unapplied_migrations):
        if int(name[:4]) > up_to:
            break

        if verbosity > 0:
            print('Applying ClickHouse migration %s for app %s in database %s' % (name, app_label, db_alias))

        migration = modules[name].Migration()
        migration.apply(db_alias, database=database)

        migration_history_model.set_migration_applied(db_alias, migrations_package, name)
        any_applied = True

    if not any_applied:
        if verbosity > 1:
            print('No migrations to apply for app "%s" does not exist' % app_label)
        return False

    return True


@receiver(post_migrate)
def clickhouse_migrate(sender, **kwargs):
    if not config.MIGRATE_WITH_DEFAULT_DB:
        # If auto migration is enabled
        return

    if kwargs.get('using', DJANGO_DEFAULT_DB_ALIAS) != DJANGO_DEFAULT_DB_ALIAS:
        # Don't call sync for every database. Just once.
        return

    app_name = kwargs['app_config'].name

    for db_alias in config.DATABASES:
        migrate_app(app_name, db_alias, verbosity=kwargs.get('verbosity', 1))


class MigrationHistory(ClickHouseModel):
    """
    A model for storing which migrations were already applied to database.
    """
    db_alias = StringField()
    package_name = StringField()
    module_name = StringField()
    applied = DateField()

    engine = MergeTree('applied', ('db_alias', 'package_name', 'module_name'))

    @classmethod
    def set_migration_applied(cls, db_alias: str, migrations_package: str, name: str) -> None:
        """
        Sets migration apply status
        :param db_alias: Database alias migration is applied to
        :param migrations_package: Package migration is stored in
        :param name: Migration name
        :return: None
        """
        # Ensure that table for migration storing is created
        for db_name in cls.migrate_non_replicated_db_aliases:
            connections[db_name].create_table(cls)

        cls.objects.create(db_alias=db_alias, package_name=migrations_package, module_name=name,
                           applied=datetime.date.today())

    @classmethod
    def get_applied_migrations(cls, db_alias: str, migrations_package: str) -> Set[str]:
        """
        Returns applied migrations names
        :param db_alias: Database alias, to check
        :param migrations_package: Package name to check
        :return: Set of migration names
        """
        qs = cls.objects.filter(package_name=migrations_package, db_alias=db_alias).only('module_name')
        try:
            return set(obj.module_name for obj in qs)
        except ServerError as ex:
            # Database doesn't exist or table doesn't exist
            # I added string check, when code parsing broke in infi.clickouse_orm
            # See https://github.com/Infinidat/infi.clickhouse_orm/issues/108
            if ex.code in {81, 60} or 'Code: 60' in ex.message or 'Code: 81,' in ex.message:
                return set()
            raise ex
        except DatabaseException as ex:
            # If database doesn't exist no migrations are applied
            # This prevents readonly=True + db_exists=False infi exception
            if str(ex) == 'Database does not exist, and cannot be created under readonly connection':
                return set()
            raise ex

    @classmethod
    def table_name(cls):
        return 'django_clickhouse_migrations'


class RunSQL(LibRunSQL):
    def __init__(self, *args, hints: Optional[dict] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.hints = hints or {}


class RunPython(LibRunPython):
    def __init__(self, *args, hints: Optional[dict] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.hints = hints or {}



import sys
from collections import namedtuple as basenamedtuple
from typing import Any, Set

from django.db import transaction, connections
from django.db.models import QuerySet


def namedtuple(*args, **kwargs):
    """
    Changes namedtuple to support defaults parameter as python 3.7 does
    https://docs.python.org/3.7/library/collections.html#collections.namedtuple
    See https://stackoverflow.com/questions/11351032/namedtuple-and-default-values-for-optional-keyword-arguments
    :return: namedtuple class
    """
    if sys.version_info < (3, 7):
        defaults = kwargs.pop('defaults', ())
        TupleClass = basenamedtuple(*args, **kwargs)
        TupleClass.__new__.__defaults__ = (None,) * (len(TupleClass._fields) - len(defaults)) + tuple(defaults)
        return TupleClass
    else:
        return basenamedtuple(*args, **kwargs)


def django_pg_returning_available(using: str) -> bool:
    """
    Checks if django-pg-returning library is installed and can be used with given databse
    :return: Boolean
    """
    try:
        import django_pg_returning  # noqa: F401
        return connections[using].vendor == 'postgresql'
    except ImportError:
        return False


def update_returning_pk(qs: QuerySet, updates: dict) -> Set[Any]:
    """
    Updates QuerySet items returning primary key values.
    This method should not depend on database engine, though can have optimization performances for some engines.
    :param qs: QuerySet to update
    :param updates: Update items as passed to QuerySet.update(**updates) method
    :return: A set of primary keys
    """
    qs._for_write = True
    if django_pg_returning_available(qs.db) and hasattr(qs, 'update_returning'):
        pk_name = qs.model._meta.pk.name
        qs = qs.only(pk_name).update_returning(**updates)
        pks = set(qs.values_list(pk_name, flat=True))
    else:
        with transaction.atomic(using=qs.db):
            pks = set(qs.select_for_update().values_list('pk', flat=True))
            QuerySet.update(qs, **updates)

    return pks



import logging
from typing import Optional, Type, Iterable

from infi.clickhouse_orm.database import Database as InfiDatabase, DatabaseException
from infi.clickhouse_orm.utils import parse_tsv
from io import BytesIO
from statsd.defaults.django import statsd

from .configuration import config
from .exceptions import DBAliasError

logger = logging.getLogger('django-clickhouse')


class Database(InfiDatabase):
    def __init__(self, **kwargs):
        infi_kwargs = {
            k: kwargs[k]
            for k in ('db_name', 'db_url', 'username', 'password', 'readonly', 'autocreate', 'timeout',
                      'verify_ssl_cert')
            if k in kwargs
        }
        super(Database, self).__init__(**infi_kwargs)

    def drop_database(self):
        # BUG fix https://github.com/Infinidat/infi.clickhouse_orm/issues/89
        super(Database, self).drop_database()
        self.db_exists = False

    def migrate(self, migrations_package_name, up_to=9999):
        raise NotImplementedError('This method is not supported by django-clickhouse.'
                                  ' Use django_clickhouse.migrations module instead.')

    def _get_applied_migrations(self, migrations_package_name):
        raise NotImplementedError("This method is not supported by django_clickhouse.")

    def select_tuples(self, query: str, model_class: Type['ClickHouseModel'],  # noqa: F821
                      settings: Optional[dict] = None) -> Iterable[tuple]:
        """
        This method selects model_class namedtuples, instead of class instances.
        Less memory consumption, greater speed
        :param query: Query to execute. Can contain substitution variables.
        :param model_class: A class of model to get fields from
        :param settings: Optional connection settings
        :return: Generator of namedtuple objects
        """
        query += ' FORMAT TabSeparatedWithNames'
        query = self._substitute(query, model_class)
        r = self._send(query, settings, True)
        lines = r.iter_lines()
        field_names = parse_tsv(next(lines))
        fields = [
            field for name, field in model_class.fields(writable=True).items()
            if name in field_names
        ]
        res_class = model_class.get_tuple_class(field_names)

        for line in lines:
            # skip blank line left by WITH TOTALS modifier
            if line:
                values = iter(parse_tsv(line))
                item = res_class(**{
                    field_name: fields[i].to_python(next(values), self.server_timezone)
                    for i, field_name in enumerate(field_names)
                })

                yield item

    def insert_tuples(self, model_class: Type['ClickHouseModel'], model_tuples: Iterable[tuple],  # noqa: F821
                      batch_size: Optional[int] = None, formatted: bool = False) -> None:
        """
        Inserts model_class namedtuples
        :param model_class: ClickHouse model, namedtuples are made from
        :param model_tuples: An iterable of tuples to insert
        :param batch_size: Size of batch
        :param formatted: If flag is set, tuples are expected to be ready to insert without calling field.to_db_string
        :return: None
        """
        tuples_iterator = iter(model_tuples)

        try:
            first_tuple = next(tuples_iterator)
        except StopIteration:
            return  # model_instances is empty

        if model_class.is_read_only() or model_class.is_system_model():
            raise DatabaseException("You can't insert into read only and system tables")

        fields_list = ','.join('`%s`' % name for name in first_tuple._fields)
        fields_dict = model_class.fields(writable=True)
        statsd_key = "%s.inserted_tuples.%s" % (config.STATSD_PREFIX, model_class.__name__)

        query = 'INSERT INTO `%s`.`%s` (%s) FORMAT TabSeparated\n' \
                % (self.db_name, model_class.table_name(), fields_list)
        query_enc = query.encode('utf-8')

        def tuple_to_csv(tup):
            if formatted:
                str_gen = (getattr(tup, field_name) for field_name in first_tuple._fields)
            else:
                str_gen = (fields_dict[field_name].to_db_string(getattr(tup, field_name), quote=False)
                           for field_name in first_tuple._fields)

            return '%s\n' % '\t'.join(str_gen)

        def gen():
            buf = BytesIO()
            buf.write(query_enc)
            buf.write(tuple_to_csv(first_tuple).encode('utf-8'))

            # Collect lines in batches of batch_size
            lines = 1
            for t in tuples_iterator:
                buf.write(tuple_to_csv(t).encode('utf-8'))

                lines += 1
                if batch_size is not None and lines >= batch_size:
                    # Return the current batch of lines
                    statsd.incr(statsd_key, lines)
                    yield buf.getvalue()
                    # Start a new batch
                    buf = BytesIO()
                    buf.write(query_enc)
                    lines = 0

            # Return any remaining lines in partial batch
            if lines:
                statsd.incr(statsd_key, lines)
                yield buf.getvalue()

        # For testing purposes
        for data in gen():
            with statsd.timer(statsd_key):
                logger.debug('django-clickhouse: insert tuple: %s' % data)
                self._send(data)


class ConnectionProxy:
    _connections = {}

    def get_connection(self, alias):
        if alias is None:
            alias = config.DEFAULT_DB_ALIAS

        if alias not in self._connections:
            if alias not in config.DATABASES:
                raise DBAliasError(alias)

            self._connections[alias] = Database(**config.DATABASES[alias])

        return self._connections[alias]

    def __getitem__(self, item):
        return self.get_connection(item)


connections = ConnectionProxy()



import datetime
import importlib
from typing import Type, Union

from celery import shared_task
from django.apps import apps as django_apps
from infi.clickhouse_orm.utils import import_submodules

from django_clickhouse.clickhouse_models import ClickHouseModel
from .configuration import config
from .utils import get_subclasses, lazy_class_import


@shared_task(queue=config.CELERY_QUEUE)
def sync_clickhouse_model(model_cls: Union[Type[ClickHouseModel], str]) -> None:
    """
    Syncs one batch of given ClickHouseModel
    :param model_cls: ClickHouseModel subclass or python path to it
    :return: None
    """
    model_cls = lazy_class_import(model_cls)

    # If sync will not finish it is not fatal to set up sync period here: sync will be executed next time
    model_cls.get_storage().set_last_sync_time(model_cls.get_import_key(), datetime.datetime.now())
    model_cls.sync_batch_from_storage()


@shared_task(queue=config.CELERY_QUEUE)
def clickhouse_auto_sync() -> None:
    """
    Plans syncing models
    :return: None
    """
    # Import all model modules
    for app in django_apps.get_app_configs():
        package_name = "%s.%s" % (app.name, config.MODELS_MODULE)
        try:
            module = importlib.import_module(package_name)
            if hasattr(module, '__path__'):
                import_submodules(package_name)
        except ImportError:
            pass

    for cls in get_subclasses(ClickHouseModel, recursive=True):
        if cls.need_sync():
            # I pass class as a string in order to make it JSON serializable
            cls_path = "%s.%s" % (cls.__module__, cls.__name__)
            sync_clickhouse_model.delay(cls_path)



"""
This file defines router to find appropriate database
"""
from typing import Type

import random
from infi.clickhouse_orm.migrations import Operation, DropTable, CreateTable

from .clickhouse_models import ClickHouseModel
from .configuration import config
from .utils import lazy_class_import


class DefaultRouter:
    def db_for_read(self, model: Type[ClickHouseModel], **hints) -> str:
        """
        Gets database to read from for model
        :param model: Model to decide for
        :param hints: Some hints to make correct choice
        :return: Database alias
        """
        return random.choice(model.read_db_aliases)

    def db_for_write(self, model: Type[ClickHouseModel], **hints) -> str:
        """
        Gets database to write to for model
        :param model: Model to decide for
        :param hints: Some hints to make correct choice
        :return: Database alias
        """
        return random.choice(model.write_db_aliases)

    def allow_migrate(self, db_alias: str, app_label: str, operation: Operation, **hints) -> bool:
        """
        Checks if migration can be applied to given database
        :param db_alias: Database alias to check
        :param app_label: App from which migration is got
        :param operation: Operation object to perform
        :param hints: Hints to make correct decision
        :return: boolean
        """
        if hints.get("force_migrate_on_databases", None):
            return db_alias in hints["force_migrate_on_databases"]

        model = hints.get('model') or getattr(operation, 'model_class', None)
        if model is None:
            raise ValueError('"model_class" attribute is not defined for operation "%s". '
                             'Please provide "force_migrate_on_databases" or "model" in hints.'
                             % operation.__class__.__name__)

        model = '%s.%s.%s' % (app_label, config.MODELS_MODULE, model) \
            if isinstance(model, str) else model
        model = lazy_class_import(model)

        if operation.__class__ not in {CreateTable, DropTable}:
            return db_alias in model.migrate_replicated_db_aliases
        else:
            return db_alias in model.migrate_non_replicated_db_aliases



from django.test import TestCase

from django_clickhouse.database import connections
from django_clickhouse.migrations import migrate_app
from tests.clickhouse_models import ClickHouseCollapseTestModel
from tests.models import TestModel


class CollapsingMergeTreeTest(TestCase):
    fixtures = ['test_model']
    maxDiff = None

    collapse_fixture = [{
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": 1,
        "version": 1
    }, {
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": -1,
        "version": 1
    }, {
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": 1,
        "version": 2
    }, {
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": -1,
        "version": 2
    }, {
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": 1,
        "version": 3
    }, {
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": -1,
        "version": 3
    }, {
        "id": 1,
        "created": "2018-01-01 00:00:00",
        "value": 0,
        "sign": 1,
        "version": 4
    }]

    def setUp(self):
        self.db = connections['default']
        self.db.drop_database()
        self.db.create_database()
        migrate_app('tests', 'default')
        ClickHouseCollapseTestModel.get_storage().flush()

        ClickHouseCollapseTestModel.objects.bulk_create([
            ClickHouseCollapseTestModel(**item) for item in self.collapse_fixture
        ])
        self.objects = TestModel.objects.filter(id=1)

    def _test_final_versions(self, final_versions):
        final_versions = list(final_versions)
        self.assertEqual(1, len(final_versions))
        item = (final_versions[0].id, final_versions[0].sign, final_versions[0].version, final_versions[0].value)
        self.assertTupleEqual((1, -1, 4, 0), item)

    def test_get_final_versions_by_final_date(self):
        final_versions = ClickHouseCollapseTestModel.engine.get_final_versions(ClickHouseCollapseTestModel,
                                                                               self.objects)
        self._test_final_versions(final_versions)

    def test_get_final_versions_by_version_date(self):
        ClickHouseCollapseTestModel.engine.version_col = 'version'
        final_versions = ClickHouseCollapseTestModel.engine.get_final_versions(ClickHouseCollapseTestModel,
                                                                               self.objects)
        self._test_final_versions(final_versions)

    def test_get_final_versions_by_final_datetime(self):
        final_versions = ClickHouseCollapseTestModel.engine.get_final_versions(ClickHouseCollapseTestModel,
                                                                               self.objects, date_col='created')
        self._test_final_versions(final_versions)

    def test_get_final_versions_by_final_no_date_col(self):
        ClickHouseCollapseTestModel.engine.date_col = None
        final_versions = ClickHouseCollapseTestModel.engine.get_final_versions(ClickHouseCollapseTestModel,
                                                                               self.objects)
        self._test_final_versions(final_versions)

    def test_get_final_versions_by_version_datetime(self):
        ClickHouseCollapseTestModel.engine.version_col = 'version'
        final_versions = ClickHouseCollapseTestModel.engine.get_final_versions(ClickHouseCollapseTestModel,
                                                                               self.objects, date_col='created')
        self._test_final_versions(final_versions)

    def test_get_final_versions_by_version_no_date_col(self):
        ClickHouseCollapseTestModel.engine.version_col = 'version'
        ClickHouseCollapseTestModel.engine.date_col = None
        final_versions = ClickHouseCollapseTestModel.engine.get_final_versions(ClickHouseCollapseTestModel,
                                                                               self.objects)
        self._test_final_versions(final_versions)

    def test_versions(self):
        ClickHouseCollapseTestModel.engine.version_col = 'version'
        batch = ClickHouseCollapseTestModel.get_insert_batch(self.objects)
        batch = list(batch)
        self.assertEqual(2, len(batch))
        self.assertEqual(4, batch[0].version)
        self.assertEqual(-1, batch[0].sign)
        self.assertEqual(5, batch[1].version)
        self.assertEqual(1, batch[1].sign)



from django.test import SimpleTestCase

from django_clickhouse.migrations import RunSQL, CreateTable
from django_clickhouse.routers import DefaultRouter
from tests.clickhouse_models import ClickHouseTestModel


class DefaultRouterAllowMigrateTest(SimpleTestCase):
    def setUp(self):
        self.router = DefaultRouter()
        self.operation = RunSQL('SELECT 1')

    def test_hints_model_class(self):
        hints = {'model': ClickHouseTestModel}

        with self.subTest('Allow migrate'):
            res = self.router.allow_migrate('default', 'tests', self.operation, **hints)
            self.assertTrue(res)

        with self.subTest('Reject migrate'):
            res = self.router.allow_migrate('other', 'tests', self.operation, **hints)
            self.assertFalse(res)

    def test_hints_model_name(self):
        hints = {'model': 'ClickHouseTestModel'}

        with self.subTest('Allow migrate'):
            res = self.router.allow_migrate('default', 'tests', self.operation, **hints)
            self.assertTrue(res)

        with self.subTest('Reject migrate'):
            res = self.router.allow_migrate('other', 'tests', self.operation, **hints)
            self.assertFalse(res)

    def test_hints_force_migrate_on_databases(self):
        hints = {'force_migrate_on_databases': ['secondary']}

        with self.subTest('Allow migrate'):
            res = self.router.allow_migrate('secondary', 'apps', self.operation, **hints)
            self.assertTrue(res)

        with self.subTest('Reject migrate'):
            res = self.router.allow_migrate('default', 'apps', self.operation, **hints)
            self.assertFalse(res)

    def test_model_operation(self):
        with self.subTest('Allow migrate'):
            operation = CreateTable(ClickHouseTestModel)
            res = self.router.allow_migrate('default', 'apps', operation)
            self.assertTrue(res)

        with self.subTest('Reject migrate'):
            operation = CreateTable(ClickHouseTestModel)
            res = self.router.allow_migrate('other', 'apps', operation)
            self.assertFalse(res)

    def test_no_model(self):
        with self.assertRaises(ValueError):
            self.router.allow_migrate('default', 'apps', self.operation)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

requires = []
with open('requirements.txt') as f:
    for line in f.readlines():
        line = line.strip()  # Remove spaces
        line = line.split('#')[0]  # Remove comments
        if line:  # Remove empty lines
            requires.append(line)

setup(
    name='django-clickhouse',
    version='1.2.2',
    packages=['django_clickhouse', 'django_clickhouse.management.commands'],
    package_dir={'': 'src'},
    url='https://github.com/carrotquest/django-clickhouse',
    license='BSD 3-clause "New" or "Revised" License',
    author='Carrot quest',
    author_email='m1ha@carrotquest.io',
    description='Django extension to integrate with ClickHouse database',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requires
)



import sys
from unittest import skipIf

from django.test import TestCase

from django_clickhouse.compatibility import namedtuple


class NamedTupleTest(TestCase):
    def test_defaults(self):
        TestTuple = namedtuple('TestTuple', ('a', 'b', 'c'), defaults=[3])
        self.assertTupleEqual((1, 2, 3), tuple(TestTuple(1, b=2)))
        self.assertTupleEqual((1, 2, 4), tuple(TestTuple(1, 2, 4)))
        self.assertTupleEqual((1, 2, 4), tuple(TestTuple(a=1, b=2, c=4)))

    @skipIf(sys.version_info < (3, 7),
            "On python < 3.7 this error is not raised, as not given defaults are filled by None")
    def test_no_required_value(self):
        TestTuple = namedtuple('TestTuple', ('a', 'b', 'c'), defaults=[3])

        with self.assertRaises(TypeError):
            TestTuple(b=1, c=4)

    def test_duplicate_value(self):
        TestTuple = namedtuple('TestTuple', ('a', 'b', 'c'), defaults=[3])

        with self.assertRaises(TypeError):
            TestTuple(1, 2, 3, c=4)

    def test_different_defaults(self):
        # Test that 2 tuple type defaults don't affect each other
        TestTuple = namedtuple('TestTuple', ('a', 'b', 'c'), defaults=[3])
        OtherTuple = namedtuple('TestTuple', ('a', 'b', 'c'), defaults=[4])
        t1 = TestTuple(a=1, b=2)
        t2 = OtherTuple(a=3, b=4)
        self.assertTupleEqual((1, 2, 3), tuple(t1))
        self.assertTupleEqual((3, 4, 4), tuple(t2))

    def test_defaults_cache(self):
        # Test that 2 tuple instances don't affect each other's defaults
        TestTuple = namedtuple('TestTuple', ('a', 'b', 'c'), defaults=[3])
        self.assertTupleEqual((1, 2, 4), tuple(TestTuple(a=1, b=2, c=4)))
        self.assertTupleEqual((1, 2, 3), tuple(TestTuple(a=1, b=2)))

    def test_equal(self):
        TestTuple = namedtuple('TestTuple', ('a', 'b', 'c'))
        t1 = TestTuple(1, 2, 3)
        t2 = TestTuple(1, 2, 3)
        self.assertEqual(t1, t2)
        self.assertEqual((1, 2, 3), t1)



"""
This file contains wrappers for infi.clckhouse_orm engines to use in django-clickhouse
"""
import datetime
import logging
from typing import List, Type, Union, Iterable, Optional, Tuple, NamedTuple

from django.db.models import Model as DjangoModel
from infi.clickhouse_orm import engines as infi_engines
from statsd.defaults.django import statsd

from .clickhouse_models import ClickHouseModel
from .configuration import config
from .database import connections
from .utils import format_datetime


logger = logging.getLogger('django-clickhouse')


class InsertOnlyEngineMixin:
    def get_insert_batch(self, model_cls: Type[ClickHouseModel], objects: List[DjangoModel]) -> Iterable[tuple]:
        """
        Gets a list of model_cls instances to insert into database
        :param model_cls: ClickHouseModel subclass to import
        :param objects: A list of django Model instances to sync
        :return: A generator of model_cls named tuples
        """
        serializer = model_cls.get_django_model_serializer(writable=True)
        return (serializer.serialize(obj) for obj in objects)


class MergeTree(InsertOnlyEngineMixin, infi_engines.MergeTree):
    pass


class ReplacingMergeTree(InsertOnlyEngineMixin, infi_engines.ReplacingMergeTree):
    pass


class SummingMergeTree(InsertOnlyEngineMixin, infi_engines.SummingMergeTree):
    pass


class CollapsingMergeTree(InsertOnlyEngineMixin, infi_engines.CollapsingMergeTree):
    pk_column = 'id'

    def __init__(self, *args, **kwargs):
        self.version_col = kwargs.pop('version_col', None)
        super(CollapsingMergeTree, self).__init__(*args, **kwargs)

    def _get_final_versions_by_version(self, db_alias: str, model_cls: Type[ClickHouseModel], object_pks: Iterable[str],
                                       columns: str, date_range_filter: str = '') -> List[NamedTuple]:
        """
        Performs request to ClickHouse in order to fetch latest version for each object pk
        :param db_alias: ClickHouse database alias used
        :param model_cls: Model class for which data is fetched
        :param object_pks: Objects primary keys to filter by
        :param columns: Columns to fetch
        :param date_range_filter: Optional date_range_filter which speeds up query if date_col is set
        :return: List of named tuples with requested columns
        """
        if date_range_filter:
            date_range_filter = 'PREWHERE {}'.format(date_range_filter)

        query = """
            SELECT {columns}
            FROM $table
            {date_range_filter}
            WHERE `{pk_column}` IN ({object_pks})
            ORDER BY `{pk_column}`, `{version_col}` DESC
            LIMIT 1 BY `{pk_column}`
        """.format(columns=','.join(columns), version_col=self.version_col, pk_column=self.pk_column,
                   date_range_filter=date_range_filter, object_pks=','.join(object_pks), sign_col=self.sign_col)

        return connections[db_alias].select_tuples(query, model_cls)

    def _get_final_versions_by_final(self, db_alias: str, model_cls: Type[ClickHouseModel], object_pks: Iterable[str],
                                     columns: str, date_range_filter: str = '') -> List[NamedTuple]:
        """
        Performs request to ClickHouse in order to fetch latest version for each object pk
        :param db_alias: ClickHouse database alias used
        :param model_cls: Model class for which data is fetched
        :param object_pks: Objects primary keys to filter by
        :param columns: Columns to fetch
        :param date_range_filter: Optional date_range_filter which speeds up query if date_col is set
        :return: List of named tuples with requested columns
        """
        if date_range_filter:
            date_range_filter += ' AND'

        query = """
            SELECT {columns} FROM $table FINAL
            WHERE {date_range_filter} `{pk_column}` IN ({object_pks})
        """
        query = query.format(columns=','.join(columns), pk_column=self.pk_column, date_range_filter=date_range_filter,
                             object_pks=','.join(object_pks))
        return connections[db_alias].select_tuples(query, model_cls)

    def _get_date_rate_filter(self, objects, model_cls: Type[ClickHouseModel], db_alias: str,
                              date_col: Optional[str]) -> str:
        """
        Generates datetime filter to speed up final queries, if date_col is present
        :param objects: Objects, which are inserted
        :param model_cls: Model class for which data is fetched
        :param db_alias: ClickHouse database alias used
        :param date_col: Optional column name, where partition date is hold. Defaults to self.date_col
        :return: String to add to WHERE or PREWHERE query section
        """
        def _dt_to_str(dt: Union[datetime.date, datetime.datetime]) -> str:
            if isinstance(dt, datetime.datetime):
                return format_datetime(dt, 0, db_alias=db_alias)
            elif isinstance(dt, datetime.date):
                return dt.isoformat()
            else:
                raise Exception('Invalid date or datetime object: `%s`' % dt)

        date_col = date_col or self.date_col

        if not date_col:
            logger.warning('django-clickhouse: date_col is not provided for model %s.'
                           ' This can cause significant performance problems while fetching data.'
                           ' It is worth inheriting CollapsingMergeTree engine with custom get_final_versions() method,'
                           ' based on your partition_key' % model_cls)
            return ''

        min_date, max_date = None, None
        for obj in objects:
            obj_date = getattr(obj, date_col)

            if min_date is None or min_date > obj_date:
                min_date = obj_date

            if max_date is None or max_date < obj_date:
                max_date = obj_date

        min_date = _dt_to_str(min_date)
        max_date = _dt_to_str(max_date)

        return "`{date_col}` >= '{min_date}' AND `{date_col}` <= '{max_date}'".\
            format(min_date=min_date, max_date=max_date, date_col=date_col)

    def get_final_versions(self, model_cls: Type[ClickHouseModel], objects: Iterable[DjangoModel],
                           date_col: Optional[str] = None) -> Iterable[tuple]:
        """
        Get objects, that are currently stored in ClickHouse.
        Depending on the partition key this can be different for different models.
        In common case, this method is optimized for date field that doesn't change.
        It also supposes primary key to by self.pk_column
        :param model_cls: ClickHouseModel subclass to import
        :param objects: Objects for which final versions are searched
        :param date_col: Optional column name, where partition date is hold. Defaults to self.date_col
        :return: A generator of named tuples, representing previous state
        """
        if not objects:
            raise StopIteration()

        object_pks = [str(getattr(obj, self.pk_column)) for obj in objects]

        db_alias = model_cls.get_database_alias()

        date_range_filter = self._get_date_rate_filter(objects, model_cls, db_alias, date_col)

        # Get fields. Sign is replaced to negative for further processing
        columns = list(model_cls.fields(writable=True).keys())
        columns.remove(self.sign_col)
        columns.append('-1 AS sign')

        params = (db_alias, model_cls, object_pks, columns, date_range_filter)

        if self.version_col:
            return self._get_final_versions_by_version(*params)
        else:
            return self._get_final_versions_by_final(*params)

    def get_insert_batch(self, model_cls: Type[ClickHouseModel], objects: List[DjangoModel]) -> Iterable[tuple]:
        """
        Gets a list of model_cls instances to insert into database
        :param model_cls: ClickHouseModel subclass to import
        :param objects: A list of django Model instances to sync
        :return: A list of model_cls objects
        """
        defaults = {self.sign_col: 1}
        if self.version_col:
            defaults[self.version_col] = 1
        serializer = model_cls.get_django_model_serializer(writable=True, defaults=defaults)
        new_objs = [serializer.serialize(obj) for obj in objects]

        statsd_key = "%s.sync.%s.steps.get_final_versions" % (config.STATSD_PREFIX, model_cls.__name__)
        with statsd.timer(statsd_key):
            # NOTE I don't use generator pattern here, as it move all time into insert.
            # That makes hard to understand where real problem is in monitoring
            old_objs = tuple(self.get_final_versions(model_cls, new_objs))

        # -1 sign has been set get_final_versions()
        old_objs_versions = {}
        for obj in old_objs:
            pk = getattr(obj, self.pk_column)
            if self.version_col:
                old_objs_versions[pk] = getattr(obj, self.version_col)
            yield obj

        # 1 sign is set by default in serializer
        for obj in new_objs:
            pk = getattr(obj, self.pk_column)
            if self.version_col:
                obj = obj._replace(**{self.version_col: old_objs_versions.get(pk, 0) + 1})

            yield obj



"""
This file defines different storages.
Storage saves intermediate data about database events - inserts, updates, delete.
This data is periodically fetched from storage and applied to ClickHouse tables.

Important:
Storage should be able to restore current importing batch, if something goes wrong.
"""
import datetime
import logging
from typing import Any, Optional, List, Tuple

import os

from celery.utils.nodenames import gethostname
from django.utils.timezone import now
from statsd.defaults.django import statsd

from .configuration import config
from .exceptions import ConfigurationError, RedisLockTimeoutError
from .redis import redis_zadd
from .utils import check_pid_exists, get_subclasses, SingletonMeta

logger = logging.getLogger('django-clickhouse')


class Storage:
    """
    Base abstract storage class, defining interface for all storages.
    The storage work algorithm:
    1) pre_sync()
    2) get_import_batch(). If batch is present go to 5)
    3) If batch is None, call get_operations()
    4) Transform operations to batch and call write_import_batch()
    5) Import batch to ClickHouse
    6) call post_sync(). If succeeded, it should remove the batch and it's data from sync_queue.

    If anything goes wrong before write_import_batch(), it is guaranteed that ClickHouse import hasn't been started yet,
    And we can repeat the procedure from the beginning.
    If anything goes wrong after write_import_batch(), we don't know it the part has been imported to ClickHouse.
    But ClickHouse is idempotent to duplicate inserts. So we can insert one batch twice correctly.
    """

    def pre_sync(self, import_key: str, **kwargs) -> None:
        """
        This method is called before import process starts
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param kwargs: Storage dependant arguments
        :return: None
        """
        pass

    def post_sync(self, import_key: str, **kwargs) -> None:
        """
        This method is called after import process has finished.
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param kwargs: Storage dependant arguments
        :return: None
        """
        pass

    def post_sync_failed(self, import_key: str, **kwargs) -> None:
        """
        This method is called after import process has finished with exception.
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param kwargs: Storage dependant arguments
        :return: None
        """
        pass

    def post_batch_removed(self, import_key: str, batch_size: int) -> None:
        """
        This method marks that batch has been removed in statsd
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param batch_size: Batch size to subtract from queue counter
        :return: None
        """
        key = "%s.sync.%s.queue" % (config.STATSD_PREFIX, import_key)
        statsd.gauge(key, self.operations_count(import_key))

    def operations_count(self, import_key: str, **kwargs) -> int:
        """
        Returns sync queue size
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param kwargs: Storage dependant arguments
        :return: Number of records in queue
        """
        raise NotImplementedError()

    def get_operations(self, import_key: str, count: int, **kwargs) -> List[Tuple[str, str]]:
        """
        Must return a list of operations on the model.
        Method should be error safe - if something goes wrong, import data should not be lost.
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param count: A batch size to get
        :param kwargs: Storage dependant arguments
        :return: A list of tuples (operation, pk) in incoming order.
        """
        raise NotImplementedError()

    def register_operations(self, import_key: str, operation: str, *pks: Any) -> int:
        """
        Registers new incoming operation
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param operation: One of insert, update, delete
        :param pk: Primary key to find records in main database. Should be string-serializable with str() method.
        :return: Number of registered operations
        """
        raise NotImplementedError()

    def register_operations_wrapped(self, import_key: str, operation: str, *pks: Any) -> int:
        """
        This is a wrapper for register_operation method, checking main parameters.
        This method should be called from inner functions.
        :param import_key: A key, returned by ClickHouseModel.get_import_key() method
        :param operation: One of insert, update, delete
        :param pks: Primary keys to find records in main database. Should be string-serializable with str() method.
        :return: Number of registered operations
        """
        if operation not in {'insert', 'update', 'delete'}:
            raise ValueError('operation must be one of [insert, update, delete]')

        statsd_key = "%s.sync.%s.register_operations" % (config.STATSD_PREFIX, import_key)
        statsd.incr(statsd_key + '.%s' % operation, len(pks))
        with statsd.timer(statsd_key):
            ops_count = self.register_operations(import_key, operation, *pks)

        statsd_key = "%s.sync.%s.queue" % (config.STATSD_PREFIX, import_key)
        statsd.gauge(statsd_key, ops_count, delta=True)
        logger.debug('django-clickhouse: registered %s on %d items (%s) to storage'
                     % (operation, len(pks), import_key))

        return ops_count

    def flush(self):
        """
        This method is used in tests to drop all storage data
        :return: None
        """
        raise NotImplementedError()

    def get_last_sync_time(self, import_key: str) -> Optional[datetime.datetime]:
        """
        Gets the last time, sync has been executed
        :return: datetime.datetime if last sync has been. Otherwise - None.
        """
        raise NotImplementedError()

    def set_last_sync_time(self, import_key: str, dt: datetime.datetime) -> None:
        """
        Sets successful sync time
        :return: None
        """
        raise NotImplementedError()


class RedisStorage(Storage, metaclass=SingletonMeta):
    """
    Fast in-memory storage made on bases of redis and redis-py library.
    Requires:
        1) REDIS database
        2) CLICKHOUSE_REDIS_CONFIG parameter defined. This should be a dict of kwargs for redis.StrictRedis(**kwargs).
    """
    REDIS_KEY_OPS_TEMPLATE = 'clickhouse_sync:operations:{import_key}'
    REDIS_KEY_RANK_TEMPLATE = 'clickhouse_sync:timstamp:{import_key}'
    REDIS_KEY_LOCK = 'clickhouse_sync:lock:{import_key}'
    REDIS_KEY_LOCK_PID = 'clickhouse_sync:lock_pid:{import_key}'
    REDIS_KEY_LAST_SYNC_TS = 'clickhouse_sync:last_sync:{import_key}'

    def __init__(self):
        # Create redis library connection. If redis is not connected properly errors should be raised
        if config.REDIS_CONFIG is None:
            raise ConfigurationError('REDIS_CONFIG')

        from redis import StrictRedis
        self._redis = StrictRedis(**config.REDIS_CONFIG)
        self._locks = {}

    def register_operations(self, import_key, operation, *pks):
        key = self.REDIS_KEY_OPS_TEMPLATE.format(import_key=import_key)
        score = datetime.datetime.now().timestamp()

        items = {'%s:%s' % (operation, str(pk)): score for pk in pks}
        return redis_zadd(self._redis, key, items)

    def operations_count(self, import_key, **kwargs):
        ops_key = self.REDIS_KEY_OPS_TEMPLATE.format(import_key=import_key)
        return self._redis.zcard(ops_key)

    def get_operations(self, import_key, count, **kwargs):
        ops_key = self.REDIS_KEY_OPS_TEMPLATE.format(import_key=import_key)
        res = self._redis.zrangebyscore(ops_key, '-inf', now().timestamp(), start=0, num=count, withscores=True)

        if res:
            ops, scores = zip(*res)

            rank_key = self.REDIS_KEY_RANK_TEMPLATE.format(import_key=import_key)
            self._redis.set(rank_key, len(ops) - 1)

            return list(tuple(op.decode().split(':')) for op in ops)
        else:
            return []

    def get_lock(self, import_key, **kwargs):
        if self._locks.get(import_key) is None:
            from .redis import RedisLock
            lock_key = self.REDIS_KEY_LOCK.format(import_key=import_key)
            lock_timeout = kwargs.get('lock_timeout', config.SYNC_DELAY * 10)
            blocking_timeout = kwargs.get('blocking_timeout', config.SYNC_DELAY)
            self._locks[import_key] = RedisLock(self._redis, lock_key, timeout=lock_timeout,
                                                blocking_timeout=blocking_timeout, thread_local=False)

        return self._locks[import_key]

    def pre_sync(self, import_key, **kwargs):
        # Block process to be single threaded. Default sync delay is 10 * default sync delay.
        # It can be changed for model, by passing `lock_timeout` argument to pre_sync
        lock = self.get_lock(import_key, **kwargs)
        current_host_name = gethostname()
        lock_pid_key = self.REDIS_KEY_LOCK_PID.format(import_key=import_key)
        try:
            lock.acquire()
            self._redis.set(lock_pid_key, '%s:%s' % (current_host_name, os.getpid()))
        except RedisLockTimeoutError:
            statsd.incr('%s.sync.%s.lock.timeout' % (config.STATSD_PREFIX, import_key))

            # Lock is busy. But If the process has been killed, I don't want to wait any more.
            # I assume that lock has been killed if it works on the same host (other than localhost)
            #   and there is no process alive.
            # I also assume that there are no hosts with same hostname other than localhost.
            # Note: previously value contained only pid. Let's support old value for back compatibility
            active_lock_data = self._redis.get(lock_pid_key).split(b":")
            active_pid = int(active_lock_data[-1] or 0)
            active_host_name = active_lock_data[0] \
                if len(active_lock_data) > 1 and active_lock_data[0] != "localhost" else None

            if (
                active_pid and active_host_name
                and active_host_name == current_host_name and not check_pid_exists(active_pid)
            ):
                statsd.incr('%s.sync.%s.lock.hard_release' % (config.STATSD_PREFIX, import_key))
                logger.warning('django-clickhouse: hard releasing lock "%s" locked by pid %d (process is dead)'
                               % (import_key, active_pid))
                self._redis.delete(lock_pid_key)
                lock.hard_release()
                self.pre_sync(import_key, **kwargs)
            else:
                raise

    def post_sync(self, import_key, **kwargs):
        rank_key = self.REDIS_KEY_RANK_TEMPLATE.format(import_key=import_key)
        ops_key = self.REDIS_KEY_OPS_TEMPLATE.format(import_key=import_key)

        top_rank = self._redis.get(rank_key)
        if top_rank:
            res = self._redis.zremrangebyrank(ops_key, 0, int(top_rank))
            batch_size = int(res)
        else:
            batch_size = 0

        # unblock lock after sync completed
        lock_pid_key = self.REDIS_KEY_LOCK_PID.format(import_key=import_key)
        self._redis.delete(lock_pid_key)
        self.post_batch_removed(import_key, batch_size)
        self.get_lock(import_key, **kwargs).release()

        logger.info('django-clickhouse: removed %d operations from storage (key: %s)' % (batch_size, import_key))

    def post_sync_failed(self, import_key, **kwargs):
        # unblock lock after sync completed
        lock_pid_key = self.REDIS_KEY_LOCK_PID.format(import_key=import_key)
        self._redis.delete(lock_pid_key)
        self.get_lock(import_key, **kwargs).release()

    def flush(self):
        key_tpls = [
            self.REDIS_KEY_RANK_TEMPLATE.format(import_key='*'),
            self.REDIS_KEY_OPS_TEMPLATE.format(import_key='*'),
            self.REDIS_KEY_LOCK.format(import_key='*'),
            self.REDIS_KEY_LAST_SYNC_TS.format(import_key='*')
        ]
        for tpl in key_tpls:
            keys = self._redis.keys(tpl)
            if keys:
                self._redis.delete(*keys)

        from .clickhouse_models import ClickHouseModel
        for model in get_subclasses(ClickHouseModel):
            if isinstance(model.get_storage(), self.__class__):
                key = "%s.sync.%s.queue" % (config.STATSD_PREFIX, model.get_import_key())
                statsd.gauge(key, 0)

    def flush_import_key(self, import_key):
        keys = [
            self.REDIS_KEY_RANK_TEMPLATE.format(import_key=import_key),
            self.REDIS_KEY_OPS_TEMPLATE.format(import_key=import_key),
            self.REDIS_KEY_LOCK.format(import_key=import_key),
            self.REDIS_KEY_LAST_SYNC_TS.format(import_key=import_key)
        ]
        self._redis.delete(*keys)
        statsd.gauge("%s.sync.%s.queue" % (config.STATSD_PREFIX, import_key), 0)

    def get_last_sync_time(self, import_key):
        sync_ts_key = self.REDIS_KEY_LAST_SYNC_TS.format(import_key=import_key)
        res = self._redis.get(sync_ts_key)
        if res is None:
            return None

        return datetime.datetime.fromtimestamp(float(res))

    def set_last_sync_time(self, import_key, dt):
        sync_ts_key = self.REDIS_KEY_LAST_SYNC_TS.format(import_key=import_key)
        self._redis.set(sync_ts_key, dt.timestamp())



import datetime
from unittest import skipIf

import django
from django.test import TransactionTestCase
from django.utils.timezone import now

from tests.clickhouse_models import ClickHouseTestModel, ClickHouseSecondTestModel, ClickHouseCollapseTestModel, \
    ClickHouseMultiTestModel
from tests.models import TestModel, SecondaryTestModel


# TestCase can't be used here:
# 1) TestCase creates transaction for inner usage
# 2) I call transaction.on_commit(), expecting no transaction at the moment
# 3) TestCase rollbacks transaction, on_commit not called
class ClickHouseDjangoModelTest(TransactionTestCase):
    def test_clickhouse_sync_models(self):
        self.assertSetEqual({ClickHouseSecondTestModel}, SecondaryTestModel.get_clickhouse_sync_models())
        self.assertSetEqual({ClickHouseTestModel, ClickHouseCollapseTestModel, ClickHouseMultiTestModel},
                            TestModel.get_clickhouse_sync_models())


class TestOperations(TransactionTestCase):
    fixtures = ['test_model']
    django_model = TestModel
    clickhouse_model = ClickHouseTestModel

    databases = ['default', 'secondary']
    db_alias = 'default'
    multi_db = True

    def setUp(self):
        self.storage = self.django_model.get_clickhouse_storage()
        self.storage.flush()
        self.before_op_items = list(self.django_model.objects.all())

    def tearDown(self):
        self.storage.flush()

    def test_save(self):
        # INSERT operation
        instance = self.django_model(created_date=datetime.date.today(), created=now(), value=2)
        instance.save()
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        # UPDATE operation
        instance.save()
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk)),
                              ('update', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_create(self):
        instance = self.django_model.objects.create(pk=100555, created_date=datetime.date.today(), created=now(),
                                                    value=2)
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_bulk_create(self):
        items = [self.django_model(created_date=datetime.date.today(), created=now(), value=i)
                 for i in range(5)]
        items = self.django_model.objects.bulk_create(items)
        self.assertEqual(5, len(items))
        self.assertSetEqual({('insert', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    @skipIf(django.VERSION < (2, 2), "bulk_update method has been introduced in django 2.2")
    def test_native_bulk_update(self):
        items = list(self.django_model.objects.filter(pk__in={1, 2}))
        for instance in items:
            instance.value = instance.pk * 10

        self.django_model.native_objects.bulk_update(items, ['value'])

        items = list(self.django_model.objects.filter(pk__in={1, 2}))
        self.assertEqual(2, len(items))
        for instance in items:
            self.assertEqual(instance.value, instance.pk * 10)

        self.assertSetEqual({('update', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_pg_bulk_create(self):
        now_dt = now()
        res = self.django_model.objects.pg_bulk_create([
            {'value': i, 'created': now_dt, 'created_date': now_dt.date()}
            for i in range(5)
        ])
        self.assertEqual(5, res)

        items = list(self.django_model.objects.filter(value__lt=100).order_by('value'))
        self.assertEqual(5, len(items))
        for i, instance in enumerate(items):
            self.assertEqual(instance.created, now_dt)
            self.assertEqual(instance.created_date, now_dt.date())
            self.assertEqual(i, instance.value)

        self.assertSetEqual({('insert', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_pg_bulk_create_returning(self):
        now_dt = now()
        res = self.django_model.objects.pg_bulk_create([
            {'value': i, 'created': now_dt, 'created_date': now_dt.date()}
            for i in range(5)
        ], returning='*')

        self.assertEqual(5, len(res))
        for i, instance in enumerate(res):
            self.assertEqual(instance.created, now_dt)
            self.assertEqual(instance.created_date, now_dt.date())
            self.assertEqual(i, instance.value)

        self.assertSetEqual({('insert', "%s.%d" % (self.db_alias, instance.pk)) for instance in res},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_pg_bulk_update(self):
        items = list(self.django_model.objects.filter(pk__in={1, 2}))

        self.django_model.objects.pg_bulk_update([
            {'id': instance.pk, 'value': instance.pk * 10}
            for instance in items
        ])

        items = list(self.django_model.objects.filter(pk__in={1, 2}))
        self.assertEqual(2, len(items))
        for instance in items:
            self.assertEqual(instance.value, instance.pk * 10)

        self.assertSetEqual({('update', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_pg_bulk_update_returning(self):
        items = list(self.django_model.objects.filter(pk__in={1, 2}))

        res = self.django_model.objects.pg_bulk_update([
            {'id': instance.pk, 'value': instance.pk * 10}
            for instance in items
        ], returning='*')

        self.assertEqual(2, len(res))
        for instance in res:
            self.assertEqual(instance.value, instance.pk * 10)

        self.assertSetEqual({('update', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_pg_bulk_update_or_create(self):
        items = list(self.django_model.objects.filter(pk__in={1, 2}))

        data = [{
            'id': instance.pk,
            'value': instance.pk * 10,
            'created_date': instance.created_date,
            'created': instance.created
        } for instance in items] + [{'id': 11, 'value': 110, 'created_date': datetime.date.today(), 'created': now()}]

        self.django_model.objects.pg_bulk_update_or_create(data)

        items = list(self.django_model.objects.filter(pk__in={1, 2, 11}))
        self.assertEqual(3, len(items))
        for instance in items:
            self.assertEqual(instance.value, instance.pk * 10)

        self.assertSetEqual({('update', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_pg_bulk_update_or_create_returning(self):
        items = list(self.django_model.objects.filter(pk__in={1, 2}))

        data = [{
            'id': instance.pk,
            'value': instance.pk * 10,
            'created_date': instance.created_date,
            'created': instance.created
        } for instance in items] + [{'id': 11, 'value': 110, 'created_date': datetime.date.today(), 'created': now()}]

        res = self.django_model.objects.pg_bulk_update_or_create(data, returning='*')

        self.assertEqual(3, len(res))
        for instance in res:
            self.assertEqual(instance.value, instance.pk * 10)

        self.assertSetEqual({('update', "%s.%d" % (self.db_alias, instance.pk)) for instance in res},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_get_or_create(self):
        instance, created = self.django_model.objects. \
            get_or_create(pk=100, defaults={'created_date': datetime.date.today(), 'created': now(),
                                            'value': 2})

        self.assertTrue(created)
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        instance, created = self.django_model.objects. \
            get_or_create(pk=100, defaults={'created_date': datetime.date.today(), 'value': 2})

        self.assertFalse(created)
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_update_or_create(self):
        instance, created = self.django_model.objects. \
            update_or_create(pk=100, defaults={'created_date': datetime.date.today(), 'created': now(), 'value': 2})
        self.assertTrue(created)
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        instance, created = self.django_model.objects. \
            update_or_create(pk=100, defaults={'created_date': datetime.date.today(), 'value': 2})

        self.assertFalse(created)
        self.assertListEqual(
            [('insert', "%s.%d" % (self.db_alias, instance.pk)), ('update', "%s.%d" % (self.db_alias, instance.pk))],
            self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_qs_update(self):
        self.django_model.objects.filter(pk=1).update(created_date=datetime.date.today())
        self.assertListEqual([('update', "%s.1" % self.db_alias)],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        self.django_model.objects.filter(created_date__lt=datetime.date.today()). \
            update(created_date=datetime.date.today())
        self.assertListEqual([('update', "%s.%d" % (self.db_alias, item.id)) for item in self.before_op_items],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_bulk_create_returning(self):
        items = [
            self.django_model(created_date=datetime.date.today(), created=now(), value=i)
            for i in range(5)
        ]
        items = self.django_model.objects.bulk_create_returning(items)
        self.assertEqual(5, len(items))
        self.assertSetEqual({('insert', "%s.%d" % (self.db_alias, instance.pk)) for instance in items},
                            set(self.storage.get_operations(self.clickhouse_model.get_import_key(), 10)))

    def test_qs_update_returning(self):
        self.django_model.objects.filter(pk=1).update_returning(created_date=datetime.date.today())
        self.assertListEqual([('update', "%s.1" % self.db_alias)],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        # Update, after which updated element will not suit update conditions
        self.django_model.objects.filter(created_date__lt=datetime.date.today()). \
            update_returning(created_date=datetime.date.today())
        self.assertListEqual([('update', "%s.%d" % (self.db_alias, item.id)) for item in self.before_op_items],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_qs_delete_returning(self):
        self.django_model.objects.filter(pk=1).delete_returning()
        self.assertListEqual([('delete', "%s.1" % self.db_alias)],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        # Delete, after which updated element will not suit update conditions
        self.django_model.objects.filter(created_date__lt=datetime.date.today()).delete_returning()
        self.assertListEqual([('delete', "%s.%d" % (self.db_alias, item.id)) for item in self.before_op_items],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_save_returning(self):
        # INSERT operation
        instance = self.django_model(created_date=datetime.date.today(), created=now(), value=2)
        instance.save_returning()
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

        # UPDATE operation
        instance.save_returning()
        self.assertListEqual([('insert', "%s.%d" % (self.db_alias, instance.pk)),
                              ('update', "%s.%d" % (self.db_alias, instance.pk))],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_delete(self):
        instance = self.django_model.objects.get(pk=1)
        instance.delete()
        self.assertListEqual([('delete', "%s.1" % self.db_alias)],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))

    def test_qs_delete(self):
        self.django_model.objects.filter(pk=1).delete()
        self.assertListEqual([('delete', "%s.1" % self.db_alias)],
                             self.storage.get_operations(self.clickhouse_model.get_import_key(), 10))


class TestSecondaryOperations(TestOperations):
    # from django.db.models.fields import *
    fixtures = ['test_secondary_model']
    django_model = SecondaryTestModel
    clickhouse_model = ClickHouseSecondTestModel
    db_alias = 'secondary'
